/*
 * Fintek F81232 USB to serial adaptor driver
 *
 * Copyright (C) 2012 Greg Kroah-Hartman (gregkh@linuxfoundation.org)
 * Copyright (C) 2012 Linux Foundation
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 as published by
 * the Free Software Foundation.
 *
 */

#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/slab.h>
#include <linux/tty.h>
#include <linux/tty_driver.h>
#include <linux/tty_flip.h>
#include <linux/serial.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/mutex.h>
#include <linux/uaccess.h>
#include <linux/usb.h>
#include <linux/usb/serial.h>
#include <linux/serial_reg.h>
#include <linux/version.h>
#include <linux/random.h>
#include <linux/gpio/driver.h>

#define VERSION "v1.16_20241219"
#define DRIVER_NAME_APPEND ""
#define MULTIDROP_ENABLE
//#define LOOPBACK
//#define DELAY_READ

/* Maximum baudrate for F81232 */
#define F81232_MAX_BAUDRATE 1500000
#define F81232_DEFAULT_BAUD_RATE 115200

/* USB Control EP parameter */
#define F81232_REGISTER_REQUEST 0xa0
#define F81232_GET_REGISTER 0xc0
#define F81232_SET_REGISTER 0x40

#define SERIAL_BASE_ADDRESS 0x0120
#define RECEIVE_BUFFER_REGISTER (0x00 + SERIAL_BASE_ADDRESS)
#define INTERRUPT_ENABLE_REGISTER (0x01 + SERIAL_BASE_ADDRESS)
#define FIFO_CONTROL_REGISTER (0x02 + SERIAL_BASE_ADDRESS)
#define LINE_CONTROL_REGISTER (0x03 + SERIAL_BASE_ADDRESS)
#define MODEM_CONTROL_REGISTER (0x04 + SERIAL_BASE_ADDRESS)
#define LINE_STATUS_REGISTER (0x05 + SERIAL_BASE_ADDRESS)
#define MODEM_STATUS_REGISTER (0x06 + SERIAL_BASE_ADDRESS)

#define MODE_CTRL_REG 0x106
#define MODE_CLK_MASK (BIT(0) | BIT(1))
#define CLKSEL_1DOT846_MHZ 0
#define CLKSEL_18DOT46_MHZ BIT(0)
#define CLKSEL_24_MHZ BIT(1)
#define CLKSEL_14DOT77_MHZ (BIT(0) | BIT(1))
#define RS485_EN BIT(2)
#define RS485_INVERT_MODE BIT(3) /* GEN2 only */

#define MODE_CONF_REG 0x107
#define TRIGGER_MULTPILE (BIT(3) | BIT(2)) /* 1x, 2x, 4x, 8x */

#ifdef MULTIDROP_ENABLE
#define ENABLE_9BITS BIT(4) /* in MODE_CTRL_REG */
#define AUTO_ADDR_9BITS BIT(5) /* in MODE_CONF_REG */
#define BADDR_EN_9BITS BIT(6) /* in MODE_CONF_REG */

#define SADDR_REG 0x108
#define SADEN_REG 0x109
#define BADDR_REG 0x10a
#endif

#define VERSION_REG 0x10f
#define VERSION_GEN1 0xff
/* Add support for RS485 invert & 9Bits/Multi-Drop mode on GEN2 */
#define VERSION_GEN2 0x01

#ifdef MULTIDROP_ENABLE
#define FINTEK_MAGIC 'F'
#define FINTEK_SET_MULTI_DROP_MODE _IOW(FINTEK_MAGIC, 0, int)
#define FINTEK_GET_MULTI_DROP_MODE _IOR(FINTEK_MAGIC, 1, int)
#define FINTEK_SET_BADDR _IOW(FINTEK_MAGIC, 6, int)
#define FINTEK_GET_BADDR _IOR(FINTEK_MAGIC, 7, int)
#endif

static const struct usb_device_id id_table[] = {
	{ USB_DEVICE(0x1934, 0x0706) },
	{} /* Terminating entry */
};
//MODULE_DEVICE_TABLE(usb, id_table);

struct f81232_private {
	struct mutex lock;
	u8 modem_control;
	u8 modem_status;
	u8 version;
	u32 current_baud_base;
	u32 multidrop_flag, baddr_flag;
	struct work_struct interrupt_work;
	struct work_struct interrupt_lsr_work;
	struct usb_serial_port *port;
	struct serial_rs485 rs485;
};

#if LINUX_VERSION_CODE < KERNEL_VERSION(4, 7, 0)
static inline bool tty_port_initialized(struct tty_port *port)
{
	return test_bit(ASYNCB_INITIALIZED, &port->flags);
	//return test_bit(ASYNCB_NORMAL_ACTIVE, &port->flags);
}
#endif

static u32 f81232_calc_baud_divisor(u32 baudrate, u32 clockrate, u32 *remain)
{
	u32 rem;

	if (!baudrate)
		return 0;

	rem = clockrate % baudrate;

	if (remain)
		*remain = rem;

	/* Round to nearest divisor */
	return DIV_ROUND_CLOSEST(clockrate, baudrate);
}

static int f81232_get_register(struct usb_serial_port *port, u16 reg, u8 *val)
{
	int status;
	u8 *tmp;
	struct usb_device *dev = port->serial->dev;

	tmp = kmalloc(sizeof(*val), GFP_KERNEL);
	if (!tmp)
		return -ENOMEM;

	status = usb_control_msg(dev, usb_rcvctrlpipe(dev, 0),
				 F81232_REGISTER_REQUEST, F81232_GET_REGISTER,
				 reg, 0, tmp, sizeof(*val),
				 /*USB_CTRL_GET_TIMEOUT*/ 9000);
	if (status != sizeof(*val)) {
		dev_err(&port->dev, "%s failed status: %d\n", __func__,
			status);

		if (status < 0)
			status = usb_translate_errors(status);
		else
			status = -EIO;
	} else {
		status = 0;
		*val = *tmp;
	}

	kfree(tmp);
	return status;
}

static int f81232_set_register(struct usb_serial_port *port, u16 reg, u8 val)
{
	int status;
	u8 *tmp;
	int retry = 10;
	struct usb_device *dev = port->serial->dev;

	tmp = kmalloc(sizeof(val), GFP_KERNEL);
	if (!tmp)
		return -ENOMEM;

	*tmp = val;

	while (--retry) {
		status = usb_control_msg(dev, usb_sndctrlpipe(dev, 0),
					 F81232_REGISTER_REQUEST,
					 F81232_SET_REGISTER, reg, 0, tmp,
					 sizeof(val),
					 /*USB_CTRL_SET_TIMEOUT*/ 9000);
		if (status == sizeof(val))
			break;
	}

	if (status != sizeof(val)) {
		dev_err(&port->dev, "%s failed status: %d\n", __func__,
			status);

		if (status < 0)
			status = usb_translate_errors(status);
		else
			status = -EIO;
	} else {
		status = 0;
	}

	kfree(tmp);
	return status;
}

static int f81232_set_mask_register(struct usb_serial_port *port, u16 reg,
				    u8 mask, u8 val)
{
	int status;
	u8 tmp;

	status = f81232_get_register(port, reg, &tmp);
	if (status)
		return status;

	tmp = (tmp & ~mask) | (val & mask);

	return f81232_set_register(port, reg, tmp);
}

static void f81232_read_msr(struct usb_serial_port *port)
{
	int status;
	u8 current_msr;
	struct tty_struct *tty;
	struct f81232_private *priv = usb_get_serial_port_data(port);

	mutex_lock(&priv->lock);
	status =
		f81232_get_register(port, MODEM_STATUS_REGISTER, &current_msr);
	if (status) {
		dev_err(&port->dev, "%s fail, status: %d\n", __func__, status);
		mutex_unlock(&priv->lock);
		return;
	}

	if (!(current_msr & UART_MSR_ANY_DELTA)) {
		mutex_unlock(&priv->lock);
		return;
	}

	priv->modem_status = current_msr;

	if (current_msr & UART_MSR_DCTS)
		port->icount.cts++;
	if (current_msr & UART_MSR_DDSR)
		port->icount.dsr++;
	if (current_msr & UART_MSR_TERI)
		port->icount.rng++;
	if (current_msr & UART_MSR_DDCD) {
		port->icount.dcd++;
		tty = tty_port_tty_get(&port->port);
		if (tty) {
			usb_serial_handle_dcd_change(
				port, tty, current_msr & UART_MSR_DCD);

			tty_kref_put(tty);
		}
	}

	wake_up_interruptible(&port->port.delta_msr_wait);
	mutex_unlock(&priv->lock);
}

static int f81232_set_mctrl(struct usb_serial_port *port, unsigned int set,
			    unsigned int clear)
{
	u8 val;
	int status;
	struct f81232_private *priv = usb_get_serial_port_data(port);

	if (((set | clear) & (TIOCM_DTR | TIOCM_RTS | TIOCM_LOOP)) == 0)
		return 0; /* no change */

	/* 'set' takes precedence over 'clear' */
	clear &= ~set;

	/* force enable interrupt with OUT2 */
	mutex_lock(&priv->lock);
	val = UART_MCR_OUT2 | priv->modem_control;

#ifdef LOOPBACK
	set |= TIOCM_LOOP;
#endif

	if (clear & TIOCM_DTR)
		val &= ~UART_MCR_DTR;

	if (clear & TIOCM_RTS)
		val &= ~UART_MCR_RTS;

	if (clear & TIOCM_LOOP)
		val &= ~UART_MCR_LOOP;

	if (set & TIOCM_DTR)
		val |= UART_MCR_DTR;

	if (set & TIOCM_RTS)
		val |= UART_MCR_RTS;

	if (set & TIOCM_LOOP)
		val |= UART_MCR_LOOP;

	dev_dbg(&port->dev, "%s new:%02x old:%02x\n", __func__, val,
		priv->modem_control);

	status = f81232_set_register(port, MODEM_CONTROL_REGISTER, val);
	if (status) {
		dev_err(&port->dev, "%s set MCR status < 0\n", __func__);
		mutex_unlock(&priv->lock);
		return status;
	}

	priv->modem_control = val;
	mutex_unlock(&priv->lock);

	return 0;
}

static void f81232_update_line_status(struct usb_serial_port *port,
				      unsigned char *data,
				      size_t actual_length)
{
	struct f81232_private *priv = usb_get_serial_port_data(port);

	if (!actual_length)
		return;

	switch (data[0] & 0x07) {
	case 0x00: /* msr change */
		dev_dbg(&port->dev, "IIR: MSR Change: %02x\n", data[0]);
		schedule_work(&priv->interrupt_work);
		break;
	case 0x02: /* tx-empty */
		break;
	case 0x04: /* rx data available */
		break;
	case 0x06: /* lsr change */
		/* we can forget it. the LSR will read from bulk-in */
		dev_dbg(&port->dev, "IIR: LSR Change: %02x\n", data[0]);
		//schedule_work(&priv->interrupt_lsr_work);
		break;
	}
}

static void f81232_read_int_callback(struct urb *urb)
{
	struct usb_serial_port *port = urb->context;
	unsigned char *data = urb->transfer_buffer;
	unsigned int actual_length = urb->actual_length;
	int status = urb->status;
	int retval;

	switch (status) {
	case 0:
		/* success */
		break;
	case -ECONNRESET:
	case -ENOENT:
	case -ESHUTDOWN:
		/* this urb is terminated, clean up */
		dev_dbg(&port->dev, "%s - urb shutting down with status: %d\n",
			__func__, status);
		return;
	default:
		dev_dbg(&port->dev, "%s - nonzero urb status received: %d\n",
			__func__, status);
		goto exit;
	}

	usb_serial_debug_data(&port->dev, __func__, urb->actual_length,
			      urb->transfer_buffer);

	f81232_update_line_status(port, data, actual_length);

exit:
	retval = usb_submit_urb(urb, GFP_ATOMIC);
	if (retval)
		dev_err(&urb->dev->dev,
			"%s - usb_submit_urb failed with result %d\n",
			__func__, retval);
}

static void f81232_process_read_urb(struct urb *urb)
{
	struct usb_serial_port *port = urb->context;
	struct f81232_private *priv = usb_get_serial_port_data(port);
	unsigned char *data = urb->transfer_buffer;
	char tty_flag;
	unsigned int i;
	u8 lsr;

	/*
	 * When opening the port we get a 1-byte packet with the current LSR,
	 * which we discard.
	 */
	if ((urb->actual_length < 2) || (urb->actual_length % 2))
		return;

	//dev_info(&port->dev, "packet size: %d\n", urb->actual_length);

	/* bulk-in data: [LSR(1Byte)+DATA(1Byte)][LSR(1Byte)+DATA(1Byte)]... */
	//dev_info(&port->dev, "urb->actual_length: %d\n", urb->actual_length);
	for (i = 0; i < urb->actual_length; i += 2) {
		tty_flag = TTY_NORMAL;
		lsr = data[i];

		if (lsr & UART_LSR_BRK_ERROR_BITS) {
			if (lsr & UART_LSR_BI) {
				tty_flag = TTY_BREAK;
				port->icount.brk++;
				usb_serial_handle_break(port);
			} else if (lsr & UART_LSR_PE) {
				tty_flag = TTY_PARITY;
				port->icount.parity++;
			} else if (lsr & UART_LSR_FE) {
				tty_flag = TTY_FRAME;
				port->icount.frame++;
			}

			if (lsr & UART_LSR_OE) {
				port->icount.overrun++;
				schedule_work(&priv->interrupt_lsr_work);
				tty_insert_flip_char(&port->port, 0,
						     TTY_OVERRUN);
			}

			dev_dbg(&port->dev, "lsr: %x\n", lsr);
		}

		if (port->port.console && port->sysrq) {
			if (usb_serial_handle_sysrq_char(port, data[i + 1]))
				continue;
		}

		tty_insert_flip_char(&port->port, data[i + 1], tty_flag);
	}

	tty_flip_buffer_push(&port->port);
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 5, 0)
static int f81232_break_ctl(struct tty_struct *tty, int break_state)
#else
static void f81232_break_ctl(struct tty_struct *tty, int break_state)
#endif
{
	/* FIXME - Stubbed out for now */

	/*
	 * break_state = -1 to turn on break, and 0 to turn off break
	 * see drivers/char/tty_io.c to see it used.
	 * last_set_data_urb_value NEVER has the break bit set in it.
	 */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 5, 0)
	return 0;
#endif
}

static int f81232_set_baudrate(struct usb_serial_port *port, speed_t baudrate)
{
	static u32 const baudrate_table[3] = { 1500000, 1152000, 921600 };
	static u8 const clock_table[3] = { CLKSEL_24_MHZ, CLKSEL_18DOT46_MHZ,
					   CLKSEL_14DOT77_MHZ };
	struct f81232_private *port_priv = usb_get_serial_port_data(port);
	u8 lcr, value;
	u32 divisor, baud_base, rem;
	int status = 0, count;

	if (baudrate <= 115200) {
		/* clock rate fixed 1.8461 MHz when below or equal 115200bps */
		value = CLKSEL_1DOT846_MHZ;
		divisor = f81232_calc_baud_divisor(baudrate, 115200, NULL);
		port_priv->current_baud_base = 115200;
	} else {
		/* Higher than 115200bps need calculate suitable clock rate */
		for (count = 0; count < ARRAY_SIZE(baudrate_table); ++count) {
			baud_base = baudrate_table[count];
			divisor = f81232_calc_baud_divisor(baudrate, baud_base,
							   &rem);
			if (!rem) {
				dev_dbg(&port->dev, "%s: found clockbase %d\n",
					__func__, baudrate_table[count]);
				value = clock_table[count];
				port_priv->current_baud_base = baud_base;
				break;
			}
		}

		if (count >= ARRAY_SIZE(baudrate_table)) {
			dev_err(&port->dev,
				"%s: cant find suitable clockbase\n",
				__func__);
			return -EINVAL;
		}
		//value = CLKSEL_18DOT46_MHZ;
	}

	//value = CLKSEL_24_MHZ;
	//divisor = 2;

	status = f81232_set_mask_register(port, MODE_CTRL_REG, MODE_CLK_MASK,
					  value);
	if (status) {
		dev_err(&port->dev, "%s failed to set CLK: %d\n", __func__,
			status);
		return -EIO;
	}

	status = f81232_get_register(port, LINE_CONTROL_REGISTER,
				     &lcr); /* get LCR */
	if (status) {
		dev_err(&port->dev, "%s failed to get LCR: %d\n", __func__,
			status);
		return -EIO;
	}

	status = f81232_set_register(port, LINE_CONTROL_REGISTER,
				     lcr | UART_LCR_DLAB); /* Enable DLAB */
	if (status) {
		dev_err(&port->dev, "%s failed to set DLAB: %d\n", __func__,
			status);
		return -EIO;
	}

	status = f81232_set_register(port, RECEIVE_BUFFER_REGISTER,
				     divisor & 0x00ff); /* low */
	if (status) {
		dev_err(&port->dev, "%s failed to set baudrate MSB: %d\n",
			__func__, status);
		goto reapply_lcr;
	}

	status = f81232_set_register(port, INTERRUPT_ENABLE_REGISTER,
				     (divisor & 0xff00) >> 8); /* high */
	if (status) {
		dev_err(&port->dev, "%s failed to set baudrate LSB: %d\n",
			__func__, status);
	}

reapply_lcr:
	status = f81232_set_register(port, LINE_CONTROL_REGISTER,
				     lcr & ~UART_LCR_DLAB);
	if (status) {
		dev_err(&port->dev, "%s failed to set DLAB: %d\n", __func__,
			status);
		return -EIO;
	}

	return 0;
}

static int f81232_port_enable(struct usb_serial_port *port)
{
	u8 val;
	int status;

	/* fifo on, trigger8, clear TX/RX*/
	val = UART_FCR_TRIGGER_8 |
	      UART_FCR_ENABLE_FIFO /*| UART_FCR_CLEAR_RCVR |
			UART_FCR_CLEAR_XMIT*/
		;

	status = f81232_set_register(port, FIFO_CONTROL_REGISTER, val);
	if (status) {
		dev_err(&port->dev, "%s failed to set FCR: %d\n", __func__,
			status);
		return status;
	}

	/* MSR Interrupt only, LSR will read from Bulk-in odd byte */
	status = f81232_set_register(port, INTERRUPT_ENABLE_REGISTER,
				     UART_IER_MSI /*| UART_IER_RLSI*/);
	if (status) {
		dev_err(&port->dev, "%s failed to set IER: %d\n", __func__,
			status);
		return status;
	}

	return 0;
}

static int f81232_port_disable(struct usb_serial_port *port)
{
	int status;

	status = f81232_set_register(port, INTERRUPT_ENABLE_REGISTER, 0);
	if (status) {
		dev_err(&port->dev, "%s failed to set IER: %d\n", __func__,
			status);
		return status;
	}

	return 0;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 1, 0)
static void f81232_set_termios(struct tty_struct *tty,
			       struct usb_serial_port *port,
			       const struct ktermios *old_termios)
#else
static void f81232_set_termios(struct tty_struct *tty,
			       struct usb_serial_port *port,
			       struct ktermios *old_termios)
#endif
{
	u8 new_lcr = 0;
	int status = 0;
	speed_t baudrate, old_baudrate = 0;

	/* Don't change anything if nothing has changed */
	if (old_termios && !tty_termios_hw_change(&tty->termios, old_termios))
		return;

	if (C_BAUD(tty) == B0)
		f81232_set_mctrl(port, 0, TIOCM_DTR | TIOCM_RTS);
	else if (old_termios && (old_termios->c_cflag & CBAUD) == B0)
		f81232_set_mctrl(port, TIOCM_DTR | TIOCM_RTS, 0);

#if 0
	baudrate = tty_get_baud_rate(tty);
	if (baudrate > 0) {
		if (baudrate > F81232_MAX_BAUDRATE) {
			baudrate = F81232_MAX_BAUDRATE;
			tty_encode_baud_rate(tty, baudrate, baudrate);
		}
		f81232_set_baudrate(port, baudrate);
	}
#else
	if (old_termios)
		old_baudrate = old_termios->c_ospeed;
	else
		old_baudrate = F81232_DEFAULT_BAUD_RATE;

	baudrate = tty_get_baud_rate(tty);
	if (baudrate) {
		/* Our device is not support for 1Mbps baudrate */
		if (baudrate == 1000000 || baudrate > F81232_MAX_BAUDRATE)
			baudrate = old_baudrate;

		if (old_baudrate != baudrate) {
			status = f81232_set_baudrate(port, baudrate);
			if (status) {
				baudrate = old_baudrate;
				f81232_set_baudrate(port, baudrate);
			}

			tty_encode_baud_rate(tty, baudrate, baudrate);
		}
	}
#endif

	if (C_PARENB(tty)) {
		new_lcr |= UART_LCR_PARITY;

		if (!C_PARODD(tty))
			new_lcr |= UART_LCR_EPAR;

		if (C_CMSPAR(tty))
			new_lcr |= UART_LCR_SPAR;
	}

	if (C_CSTOPB(tty))
		new_lcr |= UART_LCR_STOP;

	switch (C_CSIZE(tty)) {
	case CS5:
		new_lcr |= UART_LCR_WLEN5;
		break;
	case CS6:
		new_lcr |= UART_LCR_WLEN6;
		break;
	case CS7:
		new_lcr |= UART_LCR_WLEN7;
		break;
	default:
	case CS8:
		new_lcr |= UART_LCR_WLEN8;
		break;
	}

	status = f81232_set_register(port, LINE_CONTROL_REGISTER, new_lcr);
	if (status) {
		dev_err(&port->dev, "%s failed to set LCR: %d\n", __func__,
			status);
	}
}

static int f81232_tiocmget(struct tty_struct *tty)
{
	int r;
	struct usb_serial_port *port = tty->driver_data;
	struct f81232_private *port_priv = usb_get_serial_port_data(port);
	u8 mcr, msr;

	/* force get current MSR changed state */
	f81232_read_msr(port);

	mutex_lock(&port_priv->lock);
	mcr = port_priv->modem_control;
	msr = port_priv->modem_status;
	mutex_unlock(&port_priv->lock);

	r = (mcr & UART_MCR_DTR ? TIOCM_DTR : 0) |
	    (mcr & UART_MCR_RTS ? TIOCM_RTS : 0) |
	    (msr & UART_MSR_CTS ? TIOCM_CTS : 0) |
	    (msr & UART_MSR_DCD ? TIOCM_CAR : 0) |
	    (msr & UART_MSR_RI ? TIOCM_RI : 0) |
	    (msr & UART_MSR_DSR ? TIOCM_DSR : 0);

	return r;
}

static int f81232_tiocmset(struct tty_struct *tty, unsigned int set,
			   unsigned int clear)
{
	struct usb_serial_port *port = tty->driver_data;

	return f81232_set_mctrl(port, set, clear);
}

#ifdef DELAY_READ
static ssize_t f81232_submit_store(struct device *dev,
				   struct device_attribute *attr,
				   const char *buf, size_t count)
{
	struct usb_serial_port *port =
		container_of(dev, struct usb_serial_port, dev);

	usb_serial_generic_submit_read_urbs(port, GFP_KERNEL);
	printk(KERN_INFO "f81238_submit_store\n");
	return count;
}

static DEVICE_ATTR_WO(f81232_submit);
#endif

static int f81232_open(struct tty_struct *tty, struct usb_serial_port *port)
{
	int result;

#if 1
	result = f81232_port_enable(port);
	if (result)
		return result;
#endif
	/* Setup termios */
	if (tty)
		f81232_set_termios(tty, port, NULL);

	result = usb_submit_urb(port->interrupt_in_urb, GFP_KERNEL);
	if (result) {
		dev_err(&port->dev,
			"%s - failed submitting interrupt urb,"
			" error %d\n",
			__func__, result);
		return result;
	}

#ifndef DELAY_READ
	result = usb_serial_generic_open(tty, port);
	if (result) {
		usb_kill_urb(port->interrupt_in_urb);
		return result;
	}
#else
	device_create_file(&port->dev, &dev_attr_f81232_submit);
#endif

	return 0;
}

static void f81232_close(struct usb_serial_port *port)
{
	struct f81232_private *priv = usb_get_serial_port_data(port);

#if 1
	f81232_port_disable(port);
#endif
	usb_serial_generic_close(port);
	usb_kill_urb(port->interrupt_in_urb);

	flush_work(&priv->interrupt_lsr_work);
	flush_work(&priv->interrupt_work);

#ifdef DELAY_READ
	device_remove_file(&port->dev, &dev_attr_f81232_submit);
#endif
}

static void f81232_dtr_rts(struct usb_serial_port *port, int on)
{
	if (on)
		f81232_set_mctrl(port, TIOCM_DTR | TIOCM_RTS, 0);
	else
		f81232_set_mctrl(port, 0, TIOCM_DTR | TIOCM_RTS);
}

static int f81232_carrier_raised(struct usb_serial_port *port)
{
	u8 msr;
	struct f81232_private *priv = usb_get_serial_port_data(port);

	mutex_lock(&priv->lock);
	msr = priv->modem_status;
	mutex_unlock(&priv->lock);

	if (msr & UART_MSR_DCD)
		return 1;
	return 0;
}

static int f81232_get_serial_info(struct usb_serial_port *port,
				  unsigned long arg)
{
	struct f81232_private *port_priv = usb_get_serial_port_data(port);
	struct serial_struct ser;

	memset(&ser, 0, sizeof(ser));

	ser.type = PORT_16550A;
	ser.line = port->minor;
	ser.port = port->port_number;
	ser.baud_base = port_priv->current_baud_base;

	if (copy_to_user((void __user *)arg, &ser, sizeof(ser)))
		return -EFAULT;

	return 0;
}

static int f81232_ioctl_set_rs485(struct usb_serial_port *port,
				  struct serial_rs485 __user *arg)
{
	struct f81232_private *priv = usb_get_serial_port_data(port);
	struct serial_rs485 rs485;
	int status;
	u8 tmp = 0;

	status = copy_from_user(&rs485, (struct serial_rs485 __user *)arg,
				sizeof(rs485));
	if (status)
		return status;

	rs485.flags &= SER_RS485_ENABLED | SER_RS485_RTS_ON_SEND;
	if (priv->version == VERSION_GEN1 && rs485.flags == SER_RS485_ENABLED)
		return -EINVAL; /* Gen1 only support RS485 + RTS on send */

	mutex_lock(&priv->lock);
	if (rs485.flags & SER_RS485_ENABLED) {
		if (rs485.flags & SER_RS485_RTS_ON_SEND) {
			dev_dbg(&port->dev, "%s: uart_mode_rs485_1\n",
				__func__);
			tmp = RS485_EN;
		} else {
			dev_dbg(&port->dev, "%s: uart_mode_rs485\n", __func__);
			tmp = RS485_EN | RS485_INVERT_MODE;
		}
	}

	status = f81232_set_mask_register(port, MODE_CTRL_REG,
					  RS485_EN | RS485_INVERT_MODE, tmp);
	if (status) {
		mutex_unlock(&priv->lock);
		return status;
	}

	priv->rs485 = rs485;
	mutex_unlock(&priv->lock);
	return 0;
}

static int f81232_ioctl_get_rs485(struct usb_serial_port *port,
				  struct serial_rs485 __user *arg)
{
	struct f81232_private *priv = usb_get_serial_port_data(port);
	int status = 0;

	mutex_lock(&priv->lock);
	if (copy_to_user((struct serial_rs485 *)arg, &priv->rs485,
			 sizeof(priv->rs485)))
		status = -EFAULT;

	mutex_unlock(&priv->lock);
	return status;
}

#ifdef MULTIDROP_ENABLE
static int f81232_set_9bits(struct usb_serial_port *port, int __user *arg)
{
	struct f81232_private *priv = usb_get_serial_port_data(port);
	int status;
	u32 mode;
	u8 enable = 0, mask = 0, addr = 0;

	if (priv->version == VERSION_GEN1)
		return -ENOIOCTLCMD;

	if (copy_from_user(&mode, (int *)arg, sizeof(mode)))
		return -EFAULT;

	if (mode & BIT(17)) {
		enable = 1;
		addr = (mode >> 0) & 0xFF;
		mask = (mode >> 8) & 0xFF;
	}

	mutex_lock(&priv->lock);
	status = f81232_set_register(port, SADDR_REG, addr);
	if (status)
		goto end;

	status = f81232_set_register(port, SADEN_REG, mask);
	if (status)
		goto end;

	status = f81232_set_mask_register(port, INTERRUPT_ENABLE_REGISTER,
					  BIT(4), BIT(4)); /* clear SM2 */
	if (status)
		goto end;

	status = f81232_set_mask_register(port, MODE_CONF_REG, AUTO_ADDR_9BITS,
					  AUTO_ADDR_9BITS);
	if (status)
		goto end;

	status = f81232_set_mask_register(port, MODE_CTRL_REG, ENABLE_9BITS,
					  enable << 4);
	if (status)
		goto end;

	priv->multidrop_flag = mode;
end:
	mutex_unlock(&priv->lock);
	return status;
}

static int f81232_get_9bits(struct usb_serial_port *port, int __user *arg)
{
	struct f81232_private *priv = usb_get_serial_port_data(port);
	int status;

	if (priv->version == VERSION_GEN1)
		return -ENOIOCTLCMD;

	mutex_lock(&priv->lock);
	if (copy_to_user((int __user *)arg, &priv->multidrop_flag,
			 sizeof(priv->multidrop_flag)))
		status = -EFAULT;
	mutex_unlock(&priv->lock);
	return status;
}

static int f81232_set_baddr(struct usb_serial_port *port, int __user *arg)
{
	struct f81232_private *priv = usb_get_serial_port_data(port);
	int status;
	u32 mode;
	u8 enable = 0, addr = 0;

	if (priv->version == VERSION_GEN1)
		return -ENOIOCTLCMD;

	if (copy_from_user(&mode, (int *)arg, sizeof(mode)))
		return -EFAULT;

	if (mode & BIT(8)) {
		enable = 1;
		addr = (mode >> 0) & 0xFF;
	}

	mutex_lock(&priv->lock);
	status = f81232_set_register(port, BADDR_REG, addr);
	if (status)
		goto end;

	status = f81232_set_mask_register(port, INTERRUPT_ENABLE_REGISTER,
					  BIT(4), BIT(4)); /* clear SM2 */
	if (status)
		goto end;

	status = f81232_set_mask_register(port, MODE_CONF_REG, BADDR_EN_9BITS,
					  enable << 6);
	if (status)
		goto end;

	priv->baddr_flag = mode;
end:
	mutex_unlock(&priv->lock);
	return status;
}

static int f81232_get_baddr(struct usb_serial_port *port, int __user *arg)
{
	struct f81232_private *priv = usb_get_serial_port_data(port);
	int status;

	if (priv->version == VERSION_GEN1)
		return -ENOIOCTLCMD;

	mutex_lock(&priv->lock);
	if (copy_to_user((int __user *)arg, &priv->baddr_flag,
			 sizeof(priv->baddr_flag)))
		status = -EFAULT;
	mutex_unlock(&priv->lock);
	return status;
}

#endif

static int f81232_ioctl(struct tty_struct *tty, unsigned int cmd,
			unsigned long arg)
{
	struct usb_serial_port *port = tty->driver_data;

	switch (cmd) {
	case TIOCGRS485:
		return f81232_ioctl_get_rs485(
			port, (struct serial_rs485 __user *)arg);
	case TIOCSRS485:
		return f81232_ioctl_set_rs485(
			port, (struct serial_rs485 __user *)arg);
#ifdef MULTIDROP_ENABLE
	case FINTEK_SET_MULTI_DROP_MODE:
		return f81232_set_9bits(port, (int __user *)arg);
	case FINTEK_GET_MULTI_DROP_MODE:
		return f81232_get_9bits(port, (int __user *)arg);
	case FINTEK_SET_BADDR:
		return f81232_set_baddr(port, (int __user *)arg);
	case FINTEK_GET_BADDR:
		return f81232_get_baddr(port, (int __user *)arg);
#endif
	case TIOCGSERIAL:
		return f81232_get_serial_info(port, arg);
	default:
		break;
	}
	return -ENOIOCTLCMD;
}

static void f81232_interrupt_work(struct work_struct *work)
{
	struct f81232_private *priv =
		container_of(work, struct f81232_private, interrupt_work);

	f81232_read_msr(priv->port);
}

static void f81232_interrupt_lsr_work(struct work_struct *work)
{
	struct f81232_private *priv =
		container_of(work, struct f81232_private, interrupt_lsr_work);
	int status;
	u8 lsr;

	//dev_info(&priv->port->dev, "%s\n", __func__);

	status = f81232_get_register(priv->port, LINE_STATUS_REGISTER, &lsr);
	if (status) {
		dev_err(&priv->port->dev, "%s fail, status: %d\n", __func__,
			status);
	}
}

static int f81232_suspend(struct usb_serial *serial, pm_message_t message)
{
	struct usb_serial_port *port = serial->port[0];
	struct f81232_private *port_priv = usb_get_serial_port_data(port);
	int i;

	for (i = 0; i < ARRAY_SIZE(port->read_urbs); ++i)
		usb_kill_urb(port->read_urbs[i]);

	usb_kill_urb(port->interrupt_in_urb);

	if (port_priv) {
		flush_work(&port_priv->interrupt_work);
		flush_work(&port_priv->interrupt_lsr_work);
	}

	return 0;
}

static int f81232_resume(struct usb_serial *serial)
{
	struct usb_serial_port *port;
	struct f81232_private *priv;
	int status;

	port = serial->port[0];
	priv = usb_get_serial_port_data(port);

	if (tty_port_initialized(&port->port)) {
		status = usb_submit_urb(port->interrupt_in_urb, GFP_NOIO);
		if (status)
			return status;
	}

	return usb_serial_generic_resume(serial);
}

static int f81232_port_probe(struct usb_serial_port *port)
{
	struct f81232_private *priv;
	int status;

	priv = devm_kzalloc(&port->dev, sizeof(*priv), GFP_KERNEL);
	if (!priv)
		return -ENOMEM;

	mutex_init(&priv->lock);
	INIT_WORK(&priv->interrupt_work, f81232_interrupt_work);
	INIT_WORK(&priv->interrupt_lsr_work, f81232_interrupt_lsr_work);

	usb_set_serial_port_data(port, priv);

	port->port.drain_delay = 256;
	priv->port = port;

	status = f81232_get_register(port, VERSION_REG, &priv->version);
	if (status)
		return status;

	switch (priv->version) {
	case VERSION_GEN1:
		dev_info(&port->dev, "version: GEN1x probed\n");
		break;
	case VERSION_GEN2:
		dev_info(&port->dev, "version: GEN2x probed\n");
		break;
	default:
		return -ENODEV;
	}

#ifdef CONFIG_PM
	device_set_wakeup_enable(&port->serial->dev->dev, true);
#endif

	return 0;
}

static struct usb_serial_driver f81232_device = {
	.driver = {
		.owner =	THIS_MODULE,
		.name =		"f81232"DRIVER_NAME_APPEND,
	},
	.id_table =		id_table,
	.num_ports =		1,
#if 1
	.bulk_in_size =		256,
	.bulk_out_size =	256, /* 8 */
#endif
	//.bulk_out_size =	16, /* 8 */
	.open =			f81232_open,
	.close =		f81232_close,
	.dtr_rts =		f81232_dtr_rts,
	.carrier_raised =	f81232_carrier_raised,
	.ioctl =		f81232_ioctl,
	.break_ctl =		f81232_break_ctl,
	.set_termios =		f81232_set_termios,
	.tiocmget =		f81232_tiocmget,
	.tiocmset =		f81232_tiocmset,
	.tiocmiwait =		usb_serial_generic_tiocmiwait,
	.process_read_urb =	f81232_process_read_urb,
	.read_int_callback =	f81232_read_int_callback,
	.port_probe =		f81232_port_probe,
	.suspend =		f81232_suspend,
	.resume =		f81232_resume,
};

static struct usb_serial_driver *const serial_drivers[] = {
	&f81232_device,
	NULL,
};

/*
 * Fintek F81232 USB to serial adaptor driver
 *
 * Copyright (C) 2012 Greg Kroah-Hartman (gregkh@linuxfoundation.org)
 * Copyright (C) 2012 Linux Foundation
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 as published by
 * the Free Software Foundation.
 *
 */
//#undef CONFIG_GPIOLIB
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/slab.h>
#include <linux/tty.h>
#include <linux/tty_driver.h>
#include <linux/tty_flip.h>
#include <linux/serial.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/mutex.h>
#include <linux/uaccess.h>
#include <linux/usb.h>
#include <linux/usb/serial.h>
#include <linux/serial_reg.h>
#include <linux/gpio.h>
#include <linux/version.h>
#include <linux/list.h>
#include <linux/timer.h>

#define GPIO_ENABLE
#define MULTIDROP_ENABLE
#define LSR_OR_MODE
#define TEST_SEND_0_BULK 0
#define GPIO_DEFAULT_RS232 1
#define MSR_MEASURE_TIME 0
#define MSR_CALLBACK 1
#define USER_MODE

//#define DTR_ALWAYS_ON
#define GPIO_DEVICE_NODE 1
#define F81534A_CTRL_DEVICE "f81534_ctrl-%s"

#define F81534A_MAX_PORT 12
#define F81534A_CTRL_TIMER 1000
#define F81534A_CTRL_RETRY 15

/* Maximum baudrate for F81232 */
#define F81534A_MAX_BAUDRATE 3000000
#define F81534A_DEFAULT_BAUD_RATE 115200

/* USB Control EP parameter */
#define F81534A_REGISTER_REQUEST 0xa0
#define F81534A_GET_REGISTER 0xc0
#define F81534A_SET_REGISTER 0x40
#define F81534A_ACCESS_REG_RETRY 10

#define SERIAL_BASE_ADDRESS 0x0120
#define RECEIVE_BUFFER_REGISTER (0x00 + SERIAL_BASE_ADDRESS)
#define INTERRUPT_ENABLE_REGISTER (0x01 + SERIAL_BASE_ADDRESS)
#define FIFO_CONTROL_REGISTER (0x02 + SERIAL_BASE_ADDRESS)
#define LINE_CONTROL_REGISTER (0x03 + SERIAL_BASE_ADDRESS)
#define MODEM_CONTROL_REGISTER (0x04 + SERIAL_BASE_ADDRESS)
#define LINE_STATUS_REGISTER (0x05 + SERIAL_BASE_ADDRESS)
#define MODEM_STATUS_REGISTER (0x06 + SERIAL_BASE_ADDRESS)

#define MODE_CTRL_REG 0x106
#define MODE_CLK_MASK (BIT(0) | BIT(1))
#define CLKSEL_1DOT846_MHZ 0
#define CLKSEL_18DOT46_MHZ BIT(0)
#define CLKSEL_24_MHZ BIT(1)
#define CLKSEL_14DOT77_MHZ (BIT(0) | BIT(1))
#define RS485_EN BIT(2)
#define RS485_INVERT_MODE BIT(3) /* GEN2 only */

#define MODE_CONF_REG 0x107
#define TRIGGER_MASK (BIT(3) | BIT(2))
#define TRIGGER_MULTPILE_4X BIT(3)
#define TRIGGER_MULTPILE_1X 0
#define FIFO_128BYTE (BIT(1) | BIT(0))

/* FLOW_CTL_ON/OFF only valid with BIT0~6 */
#define FLOW_CTL_ON 0x10b
#define FLOW_CTL_OFF 0x10c

#define FLOW_REG 0x10d
#define HW_FLOW_CONTROL_EN BIT(0)
#define HW_FLOW_PIN_SEL BIT(1) /* Auto Flow 0:DTR/DSR 1:RTS/CTS */
#define HW_FLOW_INVERT BIT(2)
#define UART_24TO48MHZ BIT(5)
#define UART_RESET BIT(6)
#define UART_LSR_MODE BIT(7) /* 0: 1data, 1LSR , 1: ORed*/

#ifdef GPIO_ENABLE
#define GPIO_REG 0x10e
#define GPIO_SUSPEND_PULL_HIGH BIT(7)
#define GPIO_MODE2_EN BIT(6)
#define GPIO_MODE1_EN BIT(5)
#define GPIO_MODE0_EN BIT(4)
#define GPIO_SHARED BIT(3)
#define GPIO_MODE2_OUTPUT BIT(2)
#define GPIO_MODE1_OUTPUT BIT(1)
#define GPIO_MODE0_OUTPUT BIT(0)

#define GPIO_ALL_PULL_HIGH 0x07
#endif

#define GPIO_ID_PIN 0x1600
#define GPIO_CTRL_START 0x1601
#define GPIO_PULL_HIGH_PIN 0x1611
#define GPIO_CTRL_MAX_PIN 3

#define MSR_POLLING_RATE 300

#ifdef MULTIDROP_ENABLE
#define ENABLE_9BITS BIT(4) /* in MODE_CTRL_REG */
#define AUTO_ADDR_9BITS BIT(5) /* in MODE_CONF_REG */
#define BADDR_EN_9BITS BIT(6) /* in MODE_CONF_REG */

#define SADDR_REG 0x108
#define SADEN_REG 0x109
#define BADDR_REG 0x10a
#endif

#define F81534A_CMD_READ_WRITE 0x110
#define F81534A_CMD_SECTOR_ERASE 0x111

#define F81534A_CMD_ENABLE_PORT 0x116
#define F81534A_CMD_PORT_STATUS 0x117

#define F81534A_MAX_FLASH_BLOCK 64
#define F81534A_SECTOR_SIZE 0x1000
#define F81534A_CONF_SIZE 7

#define F81534A_CONF_OFFSET_GPIO 5
#define F81534A_CONF_OFFSET_UART 6

#define FINTEK_MAGIC 'F'
#define FINTEK_SET_MULTI_DROP_MODE _IOW(FINTEK_MAGIC, 0, int)
#define FINTEK_GET_MULTI_DROP_MODE _IOR(FINTEK_MAGIC, 1, int)
#define FINTEK_SET_BADDR _IOW(FINTEK_MAGIC, 6, int)
#define FINTEK_GET_BADDR _IOR(FINTEK_MAGIC, 7, int)

/* Port Control */
#define FINTEK_SET_UART_OUTPUT_MODE _IOW(FINTEK_MAGIC, 8, int)
#define FINTEK_SET_UART_OUTPUT_VAL _IOW(FINTEK_MAGIC, 9, int)
#define FINTEK_SET_UART_SUSPEND_PULL_HIGH _IOW(FINTEK_MAGIC, 10, int)
#define FINTEK_GET_UART_GPIO_MODE _IOR(FINTEK_MAGIC, 11, int)

/* H/W flow control select */
#define FINTEK_SET_UART_HWFLOW_AS_RTS _IOW(FINTEK_MAGIC, 12, int)
#define FINTEK_GET_UART_HWFLOW_AS_RTS _IOR(FINTEK_MAGIC, 13, int)

static int f81534a_disable_ports = 0;
module_param(f81534a_disable_ports, int, S_IRUGO);
MODULE_PARM_DESC(f81534a_disable_ports, "Disable f81534a serial com ports");

#if LINUX_VERSION_CODE < KERNEL_VERSION(3, 14, 0)
#define C_CMSPAR(tty) _C_FLAG((tty), CMSPAR)
#endif

static const struct usb_device_id f81534a_uart_id_table[] = {
	{ USB_DEVICE(0x2c42, 0x1602) }, /* In-Box 2 port UART device */
	{ USB_DEVICE(0x2c42, 0x1604) }, /* In-Box 4 port UART device */
	{ USB_DEVICE(0x2c42, 0x1605) }, /* In-Box 8 port UART device */
	{ USB_DEVICE(0x2c42, 0x1606) }, /* In-Box 12 port UART device */
	{ USB_DEVICE(0x2c42, 0x1608) }, /* Non-Flash type */
	{ USB_DEVICE(0x2c42, 0x1632) }, /* 2 port UART device */
	{ USB_DEVICE(0x2c42, 0x1634) }, /* 4 port UART device */
	{ USB_DEVICE(0x2c42, 0x1635) }, /* 8 port UART device */
	{ USB_DEVICE(0x2c42, 0x1636) }, /* 12 port UART device */
	{} /* Terminating entry */
};
//MODULE_DEVICE_TABLE(usb, f81534a_uart_id_table);

static const struct usb_device_id all_uart_id_table[] = {
	{ USB_DEVICE(0x1934, 0x0706) },

	{ USB_DEVICE(0x2c42, 0x1602) }, /* In-Box 2 port UART device */
	{ USB_DEVICE(0x2c42, 0x1604) }, /* In-Box 4 port UART device */
	{ USB_DEVICE(0x2c42, 0x1605) }, /* In-Box 8 port UART device */
	{ USB_DEVICE(0x2c42, 0x1606) }, /* In-Box 12 port UART device */
	{ USB_DEVICE(0x2c42, 0x1608) }, /* Non-Flash type */
	{ USB_DEVICE(0x2c42, 0x1632) }, /* 2 port UART device */
	{ USB_DEVICE(0x2c42, 0x1634) }, /* 4 port UART device */
	{ USB_DEVICE(0x2c42, 0x1635) }, /* 8 port UART device */
	{ USB_DEVICE(0x2c42, 0x1636) }, /* 12 port UART device */
	{} /* Terminating entry */
};
MODULE_DEVICE_TABLE(usb, all_uart_id_table);

static const struct usb_device_id f81534a_ctrl_id_table[] = {
	{ USB_DEVICE(0x2c42, 0x16f8) }, /* Global control device */
	{} /* Terminating entry */
};
MODULE_DEVICE_TABLE(usb, f81534a_ctrl_id_table);

struct f81534a_ctrl_private {
	struct usb_interface *intf;
	struct gpio_chip chip;
	struct mutex lock;
	struct cdev cdev;
	int device_idx;

	unsigned long long gpio_opendrain_mode;
	unsigned long long gpio_opendrain_dir;

	struct device_attribute dev_attr_gpio_operdrain_mode;
	struct device_attribute dev_attr_gpio_operdrain_dir;
	struct device_attribute dev_attr_gpio_pull_mode;
	struct device_attribute dev_attr_gpio_node;
};

struct f81534a_private {
	struct mutex lock;
	struct mutex tx_lock;

	u8 modem_control;
	u8 modem_status;
	u32 current_baud_base;
#ifdef MULTIDROP_ENABLE
	u32 multidrop_flag, baddr_flag;
#endif
	int curr_idx;
	bool has_flash;
	bool is_break;
	bool is_msr_polling;
	bool is_hwflow_use_rts;
	u8 gpio_mode;
	int tx_size, thres_on, thres_off;

	struct usb_serial_port *port;
	struct serial_rs485 rs485;
	struct delayed_work interrupt_work;
	struct work_struct interrupt_lsr_work;

#if MSR_CALLBACK
	struct urb *msr_urb;
	u8 msr_read_buffer[64];
	struct usb_ctrlrequest dr;
#endif

#if TEST_SEND_0_BULK
	u32 bulk_count;
#endif
	int update_failed_count;

	unsigned int force_baudrate;
	unsigned int current_baudrate;
	bool is_polling_msr;
	bool is_random_tx_size;
	int tx_delay;
	int tx_writer;
	int tx_delay_type;
	int tx_lsr_cnt;
	struct hrtimer tx_delay_hrt;

	struct work_struct tx_worker;
	struct delayed_work tx_no_active_work;

	unsigned int tx_kfifo_timeout;
};

struct f81534a_device {
	struct list_head list;
	struct usb_interface *intf;
	int check_index;
	int check_retry;
};

#if 0
static uint m_tx_size = 128;
static unsigned char m_thres_on = 32;
static unsigned char m_thres_off = 96;
#endif

struct delayed_work f81534a_ctrl_dwork;
static DEFINE_MUTEX(device_mutex);
static LIST_HEAD(device_list);

#if LINUX_VERSION_CODE < KERNEL_VERSION(3, 15, 0)
#define VERIFY_OCTAL_PERMISSIONS(perms)                                 \
	(BUILD_BUG_ON_ZERO((perms) < 0) +                               \
	 BUILD_BUG_ON_ZERO(                                             \
		 (perms) >                                              \
		 0777) + /* User perms >= group perms >= other perms */ \
	 BUILD_BUG_ON_ZERO(((perms) >> 6) < (((perms) >> 3) & 7)) +     \
	 BUILD_BUG_ON_ZERO((((perms) >> 3) & 7) < ((perms)&7)) + (perms))
#endif

static bool f81534a_tx_empty(struct usb_serial_port *port);
static int f81534a_submit_writer(struct usb_serial_port *port,
				 gfp_t mem_flags);

static struct f81534a_ctrl_private *
gpiochip_get_data_fintek(struct gpio_chip *chip)
{
	return container_of(chip, struct f81534a_ctrl_private, chip);
}

static u32 f81534a_calc_baud_divisor(u32 baudrate, u32 clockrate, u32 *remain)
{
	u32 rem;

	if (!baudrate)
		return 0;

	rem = clockrate % baudrate;

	if (remain)
		*remain = rem;

	/* Round to nearest divisor */
	return DIV_ROUND_CLOSEST(clockrate, baudrate);
}

static int f81534a_ctrl_get_register(struct usb_device *dev, u16 reg, u16 size,
				     void *val)
{
	int retry = F81534A_ACCESS_REG_RETRY;
	int status;
	u8 *tmp;

	tmp = kmalloc(size, GFP_KERNEL);
	if (!tmp)
		return -ENOMEM;

	while (retry--) {
		status = usb_control_msg(dev, usb_rcvctrlpipe(dev, 0),
					 F81534A_REGISTER_REQUEST,
					 F81534A_GET_REGISTER, reg, 0, tmp,
					 size, USB_CTRL_GET_TIMEOUT);
		if (status != size) {
			//pr_info("%s: status: %d, retry: %d\n", __func__, status, retry);
			status = usb_translate_errors(status);
			if (status == -EIO)
				continue;

			status = -EIO;

			usleep_range(50, 100);
		} else {
			status = 0;
			memcpy(val, tmp, size);
		}

		break;
	}

	if (status) {
		dev_err(&dev->dev, "%s: reg: %x, failed status: %d\n",
			__func__, reg, status);
	}

	kfree(tmp);
	return status;
}

static int f81534a_ctrl_set_register(struct usb_device *dev, u16 reg, u16 size,
				     void *val)
{
	int retry = F81534A_ACCESS_REG_RETRY;
	int status;
	u8 *tmp;

	tmp = kmalloc(size, GFP_KERNEL);
	if (!tmp)
		return -ENOMEM;

	memcpy(tmp, val, size);

	while (retry--) {
		status = usb_control_msg(dev, usb_sndctrlpipe(dev, 0),
					 F81534A_REGISTER_REQUEST,
					 F81534A_SET_REGISTER, reg, 0, tmp,
					 size, USB_CTRL_SET_TIMEOUT);
		if (status != size) {
			//pr_info("%s: status: %d, retry: %d\n", __func__, status, retry);
			status = usb_translate_errors(status);
			if (status == -EIO)
				continue;

			status = -EIO;
			usleep_range(50, 100);
		} else {
			status = 0;
		}

		break;
	}

	if (status) {
		dev_err(&dev->dev, "%s: reg: %x, failed status: %d\n",
			__func__, reg, status);
	}

	kfree(tmp);
	return status;
}

#if 0
static int f81534a_ctrl_get_register_raw(struct usb_device *dev, u16 reg, u16 size,
					void *val)
{
	int status;
	u8 *tmp;

	tmp = kmalloc(size, GFP_KERNEL);
	if (!tmp)
		return -ENOMEM;

	status = usb_control_msg(dev,
				usb_rcvctrlpipe(dev, 0),
				F81534A_REGISTER_REQUEST,
				F81534A_GET_REGISTER,
				reg,
				0,
				tmp,
				size,
				USB_CTRL_GET_TIMEOUT);
	if (status == size) {
		status = 0;
		memcpy(val, tmp, size);
	}

	if (status) {
		dev_err(&dev->dev, "%s: reg: %x, failed status: %d\n",
				__func__, reg, status);
	}

	kfree(tmp);
	return status;
}

static int f81534a_ctrl_set_register_raw(struct usb_device *dev, u16 reg, u16 size,
					void *val)
{
	int status;
	u8 *tmp;

	tmp = kmalloc(size, GFP_KERNEL);
	if (!tmp)
		return -ENOMEM;

	memcpy(tmp, val, size);

	status = usb_control_msg(dev,
				usb_sndctrlpipe(dev, 0),
				F81534A_REGISTER_REQUEST,
				F81534A_SET_REGISTER,
				reg,
				0,
				tmp,
				size,
				USB_CTRL_SET_TIMEOUT);
	if (status == size)
		status = 0;

	if (status) {
		dev_err(&dev->dev, "%s: reg: %x, failed status: %d\n",
				__func__, reg, status);
	}

	kfree(tmp);
	return status;
}
#endif

static int f81534a_ctrl_set_mask_register(struct usb_device *dev, u16 reg,
					  u8 mask, u8 val)
{
	int status;
	u8 tmp;

	status = f81534a_ctrl_get_register(dev, reg, 1, &tmp);
	if (status)
		return status;

	tmp = (tmp & ~mask) | (val & mask);

	status = f81534a_ctrl_set_register(dev, reg, 1, &tmp);
	if (status)
		return status;

	return 0;
}

static int f81534a_get_port_register(struct usb_serial_port *port, u16 reg,
				     u16 idx, u16 size, u8 *val)
{
	struct usb_device *dev = port->serial->dev;

	return f81534a_ctrl_get_register(dev, reg, size, val);
}

static int f81534a_set_port_register(struct usb_serial_port *port, u16 reg,
				     u16 idx, u16 size, u8 *val)
{
	struct usb_device *dev = port->serial->dev;

	return f81534a_ctrl_set_register(dev, reg, size, val);
}

static int f81534a_get_register(struct usb_serial_port *port, u16 reg, u8 *val)
{
	return f81534a_get_port_register(port, reg, 0, 1, val);
}

static int f81534a_set_register(struct usb_serial_port *port, u16 reg, u8 val)
{
	return f81534a_set_port_register(port, reg, 0, 1, &val);
}

static int f81534a_set_mask_register(struct usb_serial_port *port, u16 reg,
				     u8 mask, u8 val)
{
	int status;
	u8 tmp;

	status = f81534a_get_register(port, reg, &tmp);
	if (status)
		return status;

	tmp = (tmp & ~mask) | (val & mask);

	status = f81534a_set_register(port, reg, tmp);
	if (status)
		return status;

	return 0;
}

#if 0
static int f81534a_flash_sector_erase(struct usb_serial_port *port)
{
	return f81534a_set_port_register(port, F81534A_CMD_SECTOR_ERASE, 0, 0,
						NULL);
}

static int f81534a_flash_generic_write(struct usb_serial_port *port, u16 addr,
				u16 size, u8 *data)
{
	return f81534a_set_port_register(port, F81534A_CMD_READ_WRITE, addr,
						size, data);
}

static int f81534a_flash_write(struct usb_serial_port *port, u16 addr, u16 size,
				u8 *data)
{
	int i, status;

	for (i = 0; i < size; i += F81534A_MAX_FLASH_BLOCK) {
		status = f81534a_flash_generic_write(port, addr + i,
				min(F81534A_MAX_FLASH_BLOCK, size - i),
				&data[i]);
		if (status) {
			dev_err(&port->dev, "%s: write failed: %d\n", __func__,
					status);
			return status;
		}
	}

	return 0;
}

static int f81534a_flash_read(struct usb_serial_port *port, u16 addr, u16 size,
				u8 *data)
{
	return f81534a_get_port_register(port, F81534A_CMD_READ_WRITE, addr,
						size, data);
}
#endif

static ssize_t f81534a_reg_store(struct device *dev,
				 struct device_attribute *attr,
				 const char *buf, size_t count)
{
	struct usb_serial_port *port =
		container_of(dev, struct usb_serial_port, dev);
	int reg, data = -1;
	u8 tmp;
	u8 *p;
	int i = 0, r;

	while ((p = strsep((char **)&buf, " ")) != NULL) {
		if (!*p)
			continue;

		pr_info("%s\n", p);

		if (i == 0) {
			if (kstrtoint(p, 0, &reg) < 0)
				return -EINVAL;
		}

		if (i == 1) {
			if (kstrtoint(p, 0, &data) < 0)
				return -EINVAL;
			break;
		}

		++i;
	}

	pr_info("prepare write reg: %x, data: %x\n", reg, data);

	r = f81534a_get_register(port, reg, &tmp);
	pr_info("read reg: %x, data: %x\n", reg, tmp);

	if (data != -1) {
		tmp = data;
		r = f81534a_set_register(port, reg, tmp);
		pr_info("write reg: %x, data: %x\n", reg, tmp);
	}

	return count;
}

static ssize_t f81534a_submit_store(struct device *dev,
				    struct device_attribute *attr,
				    const char *buf, size_t count)
{
	struct usb_serial_port *port =
		container_of(dev, struct usb_serial_port, dev);
	//struct f81534a_private *priv = usb_get_serial_port_data(port);

	//usb_serial_generic_submit_read_urbs(port, GFP_KERNEL);

	dev_info(dev, "%s\n", __func__);
	f81534a_submit_writer(port, GFP_ATOMIC);

	return count;
}

static ssize_t f81534a_id_show(struct device *dev,
			       struct device_attribute *attr, char *buf)
{
	struct usb_interface *intf = to_usb_interface(dev);
	struct usb_device *usbdev = interface_to_usbdev(intf);
	uint32_t tmp;
	int status;

	status = f81534a_ctrl_get_register(usbdev, GPIO_ID_PIN, sizeof(tmp),
					   &tmp);
	if (status)
		return status;

	return sprintf(buf, "%04x\n", tmp);
}

static ssize_t f81534a_tx_size_store(struct device *dev,
				     struct device_attribute *attr,
				     const char *buf, size_t count)
{
	struct usb_serial_port *port =
		container_of(dev, struct usb_serial_port, dev);
	struct f81534a_private *priv = usb_get_serial_port_data(port);
	int data;

	if (kstrtoint(buf, 0, &data) < 0)
		return -EINVAL;

	if (data < 1 || data > 128)
		return -EINVAL;

	dev_info(dev, "old m_tx_size: %d\n", priv->tx_size);
	priv->tx_size = data;
	dev_info(dev, "new m_tx_size: %d\n", priv->tx_size);

	return count;
}

static ssize_t f81534a_thres_on_store(struct device *dev,
				      struct device_attribute *attr,
				      const char *buf, size_t count)
{
	struct usb_serial_port *port =
		container_of(dev, struct usb_serial_port, dev);
	struct f81534a_private *priv = usb_get_serial_port_data(port);
	int r, data;
	unsigned char tmp;

	if (kstrtoint(buf, 0, &data) < 0)
		return -EINVAL;

	if (data > 128)
		return -EINVAL;

	r = f81534a_get_register(port, FLOW_CTL_ON, &tmp);
	dev_info(dev, "old thres_on: %d %d\n", tmp, r);

	priv->thres_on = data;
	r = f81534a_set_register(port, FLOW_CTL_ON, priv->thres_on);
	r = f81534a_get_register(port, FLOW_CTL_ON, &tmp);
	if (priv->thres_on != tmp) {
		dev_info(dev, "new thres_on: %d %d failed!!\n", tmp, r);
		return -EINVAL;
	}

	dev_info(dev, "new thres_on: %d %d ok!!\n", tmp, r);

	return count;
}

static ssize_t f81534a_thres_off_store(struct device *dev,
				       struct device_attribute *attr,
				       const char *buf, size_t count)
{
	struct usb_serial_port *port =
		container_of(dev, struct usb_serial_port, dev);
	struct f81534a_private *priv = usb_get_serial_port_data(port);
	int r, data;
	unsigned char tmp;

	if (kstrtoint(buf, 0, &data) < 0)
		return -EINVAL;

	if (data > 128)
		return -EINVAL;

	r = f81534a_get_register(port, FLOW_CTL_OFF, &tmp);
	dev_info(dev, "old thres_on: %d %d\n", tmp, r);

	priv->thres_off = data;
	r = f81534a_set_register(port, FLOW_CTL_OFF, priv->thres_off);
	r = f81534a_get_register(port, FLOW_CTL_OFF, &tmp);
	if (priv->thres_off != tmp) {
		dev_info(dev, "new thres_off: %d %d failed !!!\n", tmp, r);
		return -EINVAL;
	}

	dev_info(dev, "new thres_off: %d %d ok\n", tmp, r);

	return count;
}

static ssize_t f81534a_force_baudrate_store(struct device *dev,
					    struct device_attribute *attr,
					    const char *buf, size_t count)
{
	struct usb_serial_port *port =
		container_of(dev, struct usb_serial_port, dev);
	struct f81534a_private *priv = usb_get_serial_port_data(port);
	int data;

	if (kstrtoint(buf, 0, &data) < 0)
		return -EINVAL;

	dev_info(dev, "old force_baudrate: %d\n", priv->force_baudrate);
	priv->force_baudrate = data;
	dev_info(dev, "new force_baudrate: %d\n", priv->force_baudrate);

	return count;
}

static ssize_t f81534a_polling_msr_store(struct device *dev,
					 struct device_attribute *attr,
					 const char *buf, size_t count)
{
	struct usb_serial_port *port =
		container_of(dev, struct usb_serial_port, dev);
	struct f81534a_private *priv = usb_get_serial_port_data(port);
	int data;

	if (kstrtoint(buf, 0, &data) < 0)
		return -EINVAL;

	dev_info(dev, "old is_polling_msr: %d\n", priv->is_polling_msr);
	priv->is_polling_msr = data;
	dev_info(dev, "new is_polling_msr: %d\n", priv->is_polling_msr);

	return count;
}

static ssize_t f81534a_tx_writer_store(struct device *dev,
				       struct device_attribute *attr,
				       const char *buf, size_t count)
{
	struct usb_serial_port *port =
		container_of(dev, struct usb_serial_port, dev);
	struct f81534a_private *priv = usb_get_serial_port_data(port);
	int data;

	if (kstrtoint(buf, 0, &data) < 0)
		return -EINVAL;

	dev_info(dev, "old tx_writer: %d\n", priv->tx_writer);
	priv->tx_writer = data;
	dev_info(dev, "new tx_writer: %d\n", priv->tx_writer);

	return count;
}

static ssize_t f81534a_tx_delay_store(struct device *dev,
				      struct device_attribute *attr,
				      const char *buf, size_t count)
{
	struct usb_serial_port *port =
		container_of(dev, struct usb_serial_port, dev);
	struct f81534a_private *priv = usb_get_serial_port_data(port);
	int data;

	if (kstrtoint(buf, 0, &data) < 0)
		return -EINVAL;

	dev_info(dev, "old tx_delay: %d\n", priv->tx_delay);
	priv->tx_delay = data;
	dev_info(dev, "new tx_delay: %d\n", priv->tx_delay);

	return count;
}

static ssize_t f81534a_tx_delay_type_store(struct device *dev,
					   struct device_attribute *attr,
					   const char *buf, size_t count)
{
	struct usb_serial_port *port =
		container_of(dev, struct usb_serial_port, dev);
	struct f81534a_private *priv = usb_get_serial_port_data(port);
	int data;

	if (kstrtoint(buf, 0, &data) < 0)
		return -EINVAL;

	dev_info(dev, "old tx_delay_type: %d\n", priv->tx_delay_type);
	priv->tx_delay_type = data;
	dev_info(dev, "new tx_delay_type: %d\n", priv->tx_delay_type);

	return count;
}

static ssize_t f81534a_tx_lsr_cnt_store(struct device *dev,
					struct device_attribute *attr,
					const char *buf, size_t count)
{
	struct usb_serial_port *port =
		container_of(dev, struct usb_serial_port, dev);
	struct f81534a_private *priv = usb_get_serial_port_data(port);
	int data;

	if (kstrtoint(buf, 0, &data) < 0)
		return -EINVAL;

	dev_info(dev, "old tx_lsr_cnt: %d\n", priv->tx_lsr_cnt);
	priv->tx_lsr_cnt = data;
	dev_info(dev, "new tx_lsr_cnt: %d\n", priv->tx_lsr_cnt);

	return count;
}

static ssize_t f81534a_tx_kfifo_store(struct device *dev,
				      struct device_attribute *attr,
				      const char *buf, size_t count)
{
	struct usb_serial_port *port =
		container_of(dev, struct usb_serial_port, dev);
	//struct f81534a_private *priv = usb_get_serial_port_data(port);

	dev_info(dev, "tx write_fifo kfifo_len: %d\n",
		 kfifo_len(&port->write_fifo));

	return count;
}

static ssize_t f81534a_tx_kfifo_timeout_store(struct device *dev,
					      struct device_attribute *attr,
					      const char *buf, size_t count)
{
	struct usb_serial_port *port =
		container_of(dev, struct usb_serial_port, dev);
	struct f81534a_private *priv = usb_get_serial_port_data(port);
	int data;

	if (kstrtoint(buf, 0, &data) < 0)
		return -EINVAL;

	dev_info(dev, "old tx_kfifo_timeout: %d\n", priv->tx_kfifo_timeout);
	priv->tx_kfifo_timeout = data;
	dev_info(dev, "new tx_kfifo_timeout: %d\n", priv->tx_kfifo_timeout);

	return count;
}

static ssize_t f81534a_tx_random_store(struct device *dev,
				       struct device_attribute *attr,
				       const char *buf, size_t count)
{
	struct usb_serial_port *port =
		container_of(dev, struct usb_serial_port, dev);
	struct f81534a_private *priv = usb_get_serial_port_data(port);
	int data;

	if (kstrtoint(buf, 0, &data) < 0)
		return -EINVAL;

	dev_info(dev, "old is_random_tx_size: %d\n", priv->is_random_tx_size);
	priv->is_random_tx_size = data;
	dev_info(dev, "new is_random_tx_size: %d\n", priv->is_random_tx_size);

	return count;
}

static DEVICE_ATTR_WO(f81534a_polling_msr);
static DEVICE_ATTR_WO(f81534a_force_baudrate);

static DEVICE_ATTR_WO(f81534a_tx_size);
static DEVICE_ATTR_WO(f81534a_tx_random);
static DEVICE_ATTR_WO(f81534a_tx_writer);
static DEVICE_ATTR_WO(f81534a_tx_delay);
static DEVICE_ATTR_WO(f81534a_tx_lsr_cnt);
static DEVICE_ATTR_WO(f81534a_tx_delay_type);
static DEVICE_ATTR_WO(f81534a_tx_kfifo);
static DEVICE_ATTR_WO(f81534a_tx_kfifo_timeout);

static DEVICE_ATTR_WO(f81534a_thres_on);
static DEVICE_ATTR_WO(f81534a_thres_off);

static DEVICE_ATTR_WO(f81534a_reg);
static DEVICE_ATTR_WO(f81534a_submit);
static DEVICE_ATTR_RO(f81534a_id);

static void f81534a_tx_no_active_work(struct work_struct *work)
{
	struct delayed_work *tx_no_active_work;
	struct f81534a_private *priv;

	tx_no_active_work = container_of(work, struct delayed_work, work);
	priv = container_of(tx_no_active_work, struct f81534a_private,
			    tx_no_active_work);

	f81534a_set_register(priv->port, (0x07 + SERIAL_BASE_ADDRESS), 0xaa);
	dev_dbg(&priv->port->dev, "%s\n", __func__);
}

static void f81534a_interrupt_lsr_work(struct work_struct *work)
{
	struct f81534a_private *priv =
		container_of(work, struct f81534a_private, interrupt_lsr_work);
	int status;
	u8 lsr;

	dev_dbg(&priv->port->dev, "%s\n", __func__);

	status = f81534a_get_register(priv->port, LINE_STATUS_REGISTER, &lsr);
	if (status) {
		dev_err(&priv->port->dev, "%s fail, status: %d\n", __func__,
			status);
	}
}

static int f81534a_prepare_write_buffer(struct usb_serial_port *port,
					void *dest, size_t size)
{
	struct f81534a_private *priv = usb_get_serial_port_data(port);
	int len, avail;
	unsigned char rng = 0;

#if TEST_SEND_0_BULK
	if (priv->bulk_count >= 100) {
		priv->bulk_count = 0;
		//dev_info(&port->dev, "zero\n");
		return 0;
	}
#endif

	// randomize tx size
	if (priv->is_random_tx_size) {
		get_random_bytes(&rng, 1);

		size = 1 + rng % 128;
		//size = 127;
	}

	avail = kfifo_len(&port->write_fifo);

	size = min_t(size_t, size, priv->tx_size);
	size = min_t(size_t, size, avail);

	if (size < 5 || size > 6) {
		//dev_info(&port->dev, "size: %ld\n", size);
	} else {
		//dev_info(&port->dev, "size got: %ld\n", size);
		size = 1;
	}

	len = kfifo_out(&port->write_fifo, dest, size);

#if TEST_SEND_0_BULK
	priv->bulk_count += len;
#endif
	//dev_info(&port->dev, "len: %d\n", len);
	return len;
}

static int f81534a_prepare_write_buffer_locked(struct usb_serial_port *port,
					       void *dest, size_t size)
{
	//struct f81534a_private *priv = usb_get_serial_port_data(port);
	unsigned long flags;
	int len;

	spin_lock_irqsave(&port->lock, flags);
	len = f81534a_prepare_write_buffer(port, dest, size);
	spin_unlock_irqrestore(&port->lock, flags);

	return len;
}

static int f81534a_submit_writer(struct usb_serial_port *port, gfp_t mem_flags)
{
	struct f81534a_private *priv = usb_get_serial_port_data(port);
	struct urb *urb;
	unsigned long flags;
	int result, size;
	struct tty_struct *tty;
	bool cts_status = true;
	speed_t baudrate;
	int tx_size = port->bulk_out_size;

	if (priv->tx_writer == 0) { // worker
		schedule_work(&priv->tx_worker);
		return 0;
	} else if (priv->tx_writer == 2) { // generic
		return usb_serial_generic_write_start(port, mem_flags);
	}

	tty = tty_port_tty_get(&port->port);
	if (tty) {
		/* check H/W Flow status */
		if (C_CRTSCTS(tty)) {
			spin_lock_irqsave(&port->lock, flags);
			if (priv->is_hwflow_use_rts)
				cts_status =
					!!(priv->modem_status & UART_MSR_CTS);
			else
				cts_status =
					!!(priv->modem_status & UART_MSR_DSR);
			spin_unlock_irqrestore(&port->lock, flags);

			baudrate = tty_get_baud_rate(tty);
			if (baudrate <= 19200)
				tx_size = 1;
		}

		tty_kref_put(tty);
	}

	if (!cts_status) {
		dev_dbg(&port->dev, "%s: CTS: %d, pause TX\n", __func__,
			cts_status);
		//dev_info(&port->dev, "%s: CTS: %d, pause TX\n", __func__, cts_status);

		return 0;
	}

	spin_lock_irqsave(&port->lock, flags);

	/* Check is any data in write_fifo */
	if (kfifo_is_empty(&port->write_fifo)) {
		spin_unlock_irqrestore(&port->lock, flags);
		usb_serial_port_softint(port);
		return 0;
	}

	/* Check H/W is TXEMPTY */
	if (!test_and_clear_bit(0, &port->write_urbs_free)) {
		spin_unlock_irqrestore(&port->lock, flags);
		return 0;
	}

	size = f81534a_prepare_write_buffer(port, port->bulk_out_buffers[0],
					    tx_size);

	//size = kfifo_out(&port->write_fifo, port->bulk_out_buffers[0],
	//		min_t(size_t, tx_size, priv->tx_size));

	//dev_info(&port->dev, "%s: tx_size: %d %d\n", __func__, tx_size, priv->tx_size);
	spin_unlock_irqrestore(&port->lock, flags);

	urb = port->write_urbs[0];
	urb->transfer_buffer_length = size;

	if (urb->transfer_buffer_length == 0 || size == 0 || tx_size == 0) {
		dev_dbg(&port->dev, "%s: urb->transfer_buffer_length: %d\n",
			__func__, urb->transfer_buffer_length);
		dev_dbg(&port->dev, "%s: size: %d\n", __func__, size);
		dev_dbg(&port->dev, "%s: tx_size: %d\n", __func__, tx_size);
	}

	result = usb_submit_urb(urb, mem_flags);
	if (result) {
		spin_lock_irqsave(&port->lock, flags);
		set_bit(0, &port->write_urbs_free);
		spin_unlock_irqrestore(&port->lock, flags);

		dev_err(&port->dev, "%s: submit failed: %d\n", __func__,
			result);
		return result;
	}

	cancel_delayed_work(&priv->tx_no_active_work);
	//flush_delayed_work(&priv->tx_no_active_work);
	schedule_delayed_work(&priv->tx_no_active_work,
			      msecs_to_jiffies(priv->tx_kfifo_timeout));

	usb_serial_port_softint(port);
	return 0;
}

static void f81534a_read_msr(struct usb_serial_port *port)
{
	int status;
	u8 current_msr;
	struct tty_struct *tty;
	struct f81534a_private *priv = usb_get_serial_port_data(port);

	status = f81534a_get_register(port, MODEM_STATUS_REGISTER,
				      &current_msr);
	if (status) {
		dev_err(&port->dev, "%s fail, status: %d\n", __func__, status);
		return;
	}

	priv->modem_status = current_msr;
	if (!(current_msr & UART_MSR_ANY_DELTA)) {
		dev_dbg(&port->dev, "%s: msr not change: %x, old: %x\n",
			__func__, current_msr, priv->modem_status);
		return;
	}

	dev_dbg(&port->dev, "%s: msr change: from %x to %x\n", __func__,
		priv->modem_status, current_msr);
	dev_dbg(&port->dev, "%s: CTS: %d\n", __func__,
		!!(current_msr & UART_MSR_CTS));

	if (current_msr & UART_MSR_DCTS) {
		port->icount.cts++;

		dev_dbg(&port->dev, "%s: DCTS detect, CTS: %d\n", __func__,
			!!(current_msr & UART_MSR_CTS));

		status = f81534a_submit_writer(port, GFP_KERNEL);
		if (status)
			dev_err(&port->dev, "%s: submit failed\n", __func__);
	}

	if (current_msr & UART_MSR_DDSR)
		port->icount.dsr++;
	if (current_msr & UART_MSR_TERI)
		port->icount.rng++;
	if (current_msr & UART_MSR_DDCD) {
		port->icount.dcd++;
		tty = tty_port_tty_get(&port->port);
		if (tty) {
			usb_serial_handle_dcd_change(
				port, tty, current_msr & UART_MSR_DCD);

			tty_kref_put(tty);
		}
	}

	wake_up_interruptible(&port->port.delta_msr_wait);
}

static int f81534a_set_mctrl(struct usb_serial_port *port, unsigned int set,
			     unsigned int clear)
{
	u8 val;
	int status;
	struct f81534a_private *priv = usb_get_serial_port_data(port);

	if (((set | clear) & (TIOCM_DTR | TIOCM_RTS | TIOCM_LOOP)) == 0)
		return 0; /* no change */

	/* 'set' takes precedence over 'clear' */
	clear &= ~set;

	/* force enable interrupt with OUT2 */
	if (mutex_lock_interruptible(&priv->lock))
		return -EINTR;

	val = UART_MCR_OUT2 | priv->modem_control;

#ifdef LOOPBACK
	set |= TIOCM_LOOP;
#endif

	if (set & TIOCM_LOOP)
		dev_dbg(&port->dev, "Internal Loopback Enabled\n");

	if (clear & TIOCM_DTR) {
		//dev_info(&port->dev, "clear DTR\n");
		val &= ~UART_MCR_DTR;
	}

	if (clear & TIOCM_RTS) {
		//dev_info(&port->dev, "clear RTS\n");
		val &= ~UART_MCR_RTS;
	}

	if (clear & TIOCM_LOOP) {
		//dev_info(&port->dev, "clear LOOP\n");
		val &= ~UART_MCR_LOOP;
	}

	if (set & TIOCM_DTR) {
		//dev_info(&port->dev, "set DTR\n");
		val |= UART_MCR_DTR;
	}

	if (set & TIOCM_RTS) {
		//dev_info(&port->dev, "set RTS\n");
		val |= UART_MCR_RTS;
	}

	if (set & TIOCM_LOOP) {
		//dev_info(&port->dev, "set LOOP\n");
		val |= UART_MCR_LOOP;
	}

#ifdef DTR_ALWAYS_ON
	val &= ~UART_MCR_DTR;
#endif

	dev_dbg(&port->dev, "%s new:%02x old:%02x\n", __func__, val,
		priv->modem_control);

	status = f81534a_set_register(port, MODEM_CONTROL_REGISTER, val);
	if (status) {
		dev_err(&port->dev, "%s set MCR status < 0\n", __func__);
		mutex_unlock(&priv->lock);
		return status;
	}

	priv->modem_control = val;
	//usleep_range(200,2000);
	mutex_unlock(&priv->lock);

	return 0;
}

static void f81534a_update_line_status(struct usb_serial_port *port,
				       unsigned char *data,
				       size_t actual_length)
{
	struct f81534a_private *priv = usb_get_serial_port_data(port);

	if (data[0] & UART_IIR_NO_INT)
		return;

	switch (data[0] & UART_IIR_ID) {
	case UART_IIR_MSI: /* msr change */
		dev_dbg(&port->dev,
			"IIR: MSR Change: %02x %02x %02x %02x %02x %02x %02x\n",
			data[0], data[1], data[2], data[3], data[4], data[5],
			data[6]);
		schedule_delayed_work(&priv->interrupt_work, 0);
		break;
	case UART_IIR_THRI: /* tx-empty */
		break;
	case UART_IIR_RDI: /* rx data available */
		break;
	case UART_IIR_RLSI: /* lsr change */
		/* we can forget it. the LSR will read from bulk-in */
		dev_dbg(&port->dev, "IIR: LSR Change: %02x\n", data[0]);
		break;
	}

	return;
}

static void f81534a_read_int_callback(struct urb *urb)
{
	struct usb_serial_port *port = urb->context;
	unsigned char *data = urb->transfer_buffer;
	unsigned int actual_length = urb->actual_length;
	int status = urb->status;
	int retval;

	switch (status) {
	case 0:
		/* success */
		break;
	case -ECONNRESET:
	case -ENOENT:
	case -ESHUTDOWN:
		/* this urb is terminated, clean up */
		dev_dbg(&port->dev, "%s - urb shutting down with status: %d\n",
			__func__, status);
		return;
	default:
		dev_dbg(&port->dev, "%s - nonzero urb status received: %d\n",
			__func__, status);
		goto exit;
	}

	usb_serial_debug_data(&port->dev, __func__, urb->actual_length,
			      urb->transfer_buffer);

	f81534a_update_line_status(port, data, actual_length);

exit:
	retval = usb_submit_urb(urb, GFP_ATOMIC);
	if (retval)
		dev_err(&urb->dev->dev,
			"%s - usb_submit_urb failed with result %d\n",
			__func__, retval);
}

static void f81534a_process_read_urb(struct urb *urb)
{
	struct usb_serial_port *port = urb->context;
	struct f81534a_private *priv = usb_get_serial_port_data(port);
	unsigned char *data = urb->transfer_buffer;
	unsigned int i;
	u8 tty_flag;
	u8 lsr;
	u8 len;

	if (urb->status) {
		dev_err(&port->dev, "%s: urb->status: %d\n", __func__,
			urb->status);
		pr_err("%s: urb->status: %d\n", __func__, urb->status);
	}

	if (urb->actual_length < 2) {
		dev_err(&port->dev, "%s: urb->actual_length(%d) < 2\n",
			__func__, urb->actual_length);
		return;
	}

	len = data[0];

	if (len != urb->actual_length) {
		dev_err(&port->dev, "%s: len(%d) != urb->actual_length(%d)\n",
			__func__, len, urb->actual_length);
		pr_err("%s: len(%d) != urb->actual_length(%d)\n", __func__,
		       len, urb->actual_length);
		return;
	}

	//dev_info(&port->dev, "%s: urb->actual_length == %d\n", __func__, urb->actual_length);

#ifdef LSR_OR_MODE
	lsr = data[len - 1];
	tty_flag = TTY_NORMAL;

	if (lsr & UART_LSR_BRK_ERROR_BITS) {
		dev_dbg(&port->dev, "%s: lsr: %x, len: %d\n", __func__, lsr,
			urb->actual_length);

		//pr_err("%s: lsr: %x, len: %d\n", __func__, lsr, urb->actual_length);

		if (lsr & UART_LSR_BI) {
			dev_dbg(&port->dev, "UART_LSR_BI,");
			//pr_err("UART_LSR_BI,");
		}

		if (lsr & UART_LSR_PE) {
			dev_dbg(&port->dev, "UART_LSR_PE,");
			//pr_err("UART_LSR_PE,");
		}

		if (lsr & UART_LSR_FE) {
			dev_dbg(&port->dev, "UART_LSR_FE,");
			//pr_err("UART_LSR_FE,");
		}

		if (lsr & UART_LSR_OE) {
			dev_dbg(&port->dev, "UART_LSR_OE,");
			//pr_err("UART_LSR_OE,");
		}

		dev_dbg(&port->dev, "\n");

		if (lsr & UART_LSR_BI) {
			tty_flag = TTY_BREAK;
			port->icount.brk++;
			usb_serial_handle_break(port);
		} else if (lsr & UART_LSR_PE) {
			tty_flag = TTY_PARITY;
			port->icount.parity++;
		} else if (lsr & UART_LSR_FE) {
			tty_flag = TTY_FRAME;
			port->icount.frame++;
		}

		if (lsr & UART_LSR_OE) {
			port->icount.overrun++;
			schedule_work(&priv->interrupt_lsr_work);
			tty_insert_flip_char(&port->port, 0, TTY_OVERRUN);
		}
	}

	for (i = 1; i < urb->actual_length - 1; ++i) {
		if (port->port.console && port->sysrq) {
			if (usb_serial_handle_sysrq_char(port, data[i]))
				continue;
		}

		//dev_dbg(&port->dev, "%s: i: %d, %x\n", __func__, i, data[i]);

		tty_insert_flip_char(&port->port, data[i], tty_flag);
	}
#else
	for (i = 1; i < urb->actual_length - 1; i += 2) {
		lsr = data[i];

		if (lsr & UART_LSR_BRK_ERROR_BITS) {
			dev_dbg(&port->dev, "%s: lsr: %x, len: %d\n", __func__,
				lsr, urb->actual_length);

			if (lsr & UART_LSR_BI)
				dev_dbg(&port->dev, "UART_LSR_BI,");

			if (lsr & UART_LSR_PE)
				dev_dbg(&port->dev, "UART_LSR_PE,");

			if (lsr & UART_LSR_FE)
				dev_dbg(&port->dev, "UART_LSR_FE,");

			if (lsr & UART_LSR_OE)
				dev_dbg(&port->dev, "UART_LSR_OE,");

			dev_dbg(&port->dev, "\n");

			if (lsr & UART_LSR_BI) {
				tty_flag = TTY_BREAK;
				port->icount.brk++;
				usb_serial_handle_break(port);
			} else if (lsr & UART_LSR_PE) {
				tty_flag = TTY_PARITY;
				port->icount.parity++;
			} else if (lsr & UART_LSR_FE) {
				tty_flag = TTY_FRAME;
				port->icount.frame++;
			}

			if (lsr & UART_LSR_OE) {
				port->icount.overrun++;
				schedule_work(&priv->interrupt_lsr_work);
				tty_insert_flip_char(&port->port, 0,
						     TTY_OVERRUN);
			}
		}

		if (port->port.console && port->sysrq) {
			if (usb_serial_handle_sysrq_char(port, data[i + 1]))
				continue;
		}

		tty_insert_flip_char(&port->port, data[i + 1], TTY_NORMAL);
	}
#endif
	tty_flip_buffer_push(&port->port);
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 5, 0)
static int f81534a_break_ctl(struct tty_struct *tty, int break_state)
#else
static void f81534a_break_ctl(struct tty_struct *tty, int break_state)
#endif
{
	struct usb_serial_port *port = tty->driver_data;
	struct f81534a_private *priv = usb_get_serial_port_data(port);
	int status;
	u8 value = 0;

	mutex_lock(&priv->lock);

	priv->is_break = !!break_state;
	if (priv->is_break)
		value = UART_LCR_SBC;

	status = f81534a_set_mask_register(port, LINE_CONTROL_REGISTER,
					   UART_LCR_SBC, value);
	if (status)
		dev_err(&port->dev, "set break failed: %d\n", status);

	mutex_unlock(&priv->lock);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 5, 0)
	return status;
#endif
}

static int f81534a_set_baudrate(struct usb_serial_port *port, speed_t baudrate)
{
	static u32 const baudrate_table[] = { 115200, 921600, 1152000, 1500000,
					      3000000 };
	static u8 const clock_table[] = { CLKSEL_1DOT846_MHZ,
					  CLKSEL_14DOT77_MHZ,
					  CLKSEL_18DOT46_MHZ, CLKSEL_24_MHZ,
					  CLKSEL_24_MHZ };

	struct f81534a_private *priv = usb_get_serial_port_data(port);
	u8 value, tmp = 0;
	u32 divisor, baud_base, rem;
	int status = 0, count = 0;

	dev_dbg(&port->dev, "%s: set baudrate %d\n", __func__, baudrate);

	/* calculate suitable clock rate */
	for (count = 0; count < ARRAY_SIZE(baudrate_table); ++count) {
		baud_base = baudrate_table[count];
		divisor = f81534a_calc_baud_divisor(baudrate, baud_base, &rem);
		if (!rem) {
			dev_dbg(&port->dev,
				"%s: found clockbase %d, divisor: %d\n",
				__func__, baudrate_table[count], divisor);
			value = clock_table[count];
			priv->current_baud_base = baud_base;
			break;
		}
	}

	if (count >= ARRAY_SIZE(baudrate_table)) {
		if (baudrate <= 115200) {
			count = 0;
			value = clock_table[count];
			priv->current_baud_base = baudrate_table[count];
			divisor = DIV_ROUND_CLOSEST(baudrate_table[count],
						    baudrate);
			dev_dbg(&port->dev, "%s: special baudrate %d\n",
				__func__, baudrate);
		} else {
			dev_err(&port->dev,
				"%s: cant find suitable clockbase\n",
				__func__);
			return -EINVAL;
		}
	}

#if 0
	count = 0;
	value = CLKSEL_24_MHZ;
	divisor = 1;
#endif

	if (count == ARRAY_SIZE(clock_table) - 1)
		tmp = UART_24TO48MHZ;

	/* Check for using 24 to 48MHz multipler */
	status =
		f81534a_set_mask_register(port, FLOW_REG, UART_24TO48MHZ, tmp);
	if (status) {
		dev_err(&port->dev, "%s failed to set FLOW_REG: %d\n",
			__func__, status);
		return status;
	}

	status = f81534a_set_mask_register(port, MODE_CTRL_REG, MODE_CLK_MASK,
					   value);
	if (status) {
		dev_err(&port->dev, "%s failed to set CLK: %d\n", __func__,
			status);
		return status;
	}

	status = f81534a_set_register(port, LINE_CONTROL_REGISTER,
				      UART_LCR_DLAB); /* Enable DLAB */
	if (status) {
		dev_err(&port->dev, "%s failed to set DLAB: %d\n", __func__,
			status);
		return status;
	}

	status = f81534a_set_register(port, RECEIVE_BUFFER_REGISTER,
				      divisor & 0x00ff); /* low */
	if (status) {
		dev_err(&port->dev, "%s failed to set baudrate MSB: %d\n",
			__func__, status);
		return status;
	}

	status = f81534a_set_register(port, INTERRUPT_ENABLE_REGISTER,
				      (divisor & 0xff00) >> 8); /* high */
	if (status) {
		dev_err(&port->dev, "%s failed to set baudrate LSB: %d\n",
			__func__, status);
		return status;
	}

	return 0;
}

static int f81534a_port_enable(struct usb_serial_port *port)
{
	u8 val;
	int status;

#if 0
	status = f81534a_get_register(port, FLOW_REG, &val);
	if (status) {
		dev_err(&port->dev, "%s failed to get FLOW_REG: %d\n",
			__func__, status);
		return status;
	}

	//dev_info(&port->dev, "FLOW_REG: %x\n", val);
#endif

#if 0
	status = f81534a_set_mask_register(port, FLOW_REG, UART_RESET,
			UART_RESET);
	if (status) {
		dev_err(&port->dev, "%s failed to set FLOW_REG: %d\n",
			__func__, status);
		return status;
	}
#endif

#ifdef LSR_OR_MODE
	val = UART_LSR_MODE;
	//dev_info(&port->dev, "lsr or\n");
#else
	val = 0;
	//dev_info(&port->dev, "lsr all\n");
#endif

	/* LSR Mode */
	status = f81534a_set_mask_register(port, FLOW_REG, UART_LSR_MODE, val);
	if (status) {
		dev_err(&port->dev, "%s failed to set LSR Mode: %d\n",
			__func__, status);
		return status;
	}

	val = TRIGGER_MULTPILE_4X | FIFO_128BYTE;
	status = f81534a_set_mask_register(port, MODE_CONF_REG,
					   TRIGGER_MASK | FIFO_128BYTE, val);
	if (status) {
		dev_err(&port->dev, "%s failed to set MODE_CONF_REG: %d\n",
			__func__, status);
		return status;
	}

	/* fifo on, trigger8, clear TX/RX */
	val = UART_FCR_TRIGGER_8 | UART_FCR_ENABLE_FIFO | UART_FCR_CLEAR_RCVR |
	      UART_FCR_CLEAR_XMIT;

	status = f81534a_set_register(port, FIFO_CONTROL_REGISTER, val);
	if (status) {
		dev_err(&port->dev, "%s failed to set FCR: %d\n", __func__,
			status);
		return status;
	}

#if 1
	val = 0;
#else
	val = UART_IER_MSI;
#endif
	/* MSR Interrupt only, LSR will read from Bulk-in odd byte */
	status = f81534a_set_register(port, INTERRUPT_ENABLE_REGISTER, val);
	if (status) {
		dev_err(&port->dev, "%s failed to set IER: %d\n", __func__,
			status);
		return status;
	}

	return 0;
}

static int f81534a_port_disable(struct usb_serial_port *port)
{
	int status;

	status = f81534a_set_register(port, INTERRUPT_ENABLE_REGISTER, 0);
	if (status) {
		dev_err(&port->dev, "%s failed to set IER: %d\n", __func__,
			status);
		return status;
	}

	return 0;
}

static void f81534a_set_autoflow_mode(struct usb_serial_port *port,
				      speed_t baud, bool en)
{
	struct f81534a_private *priv = usb_get_serial_port_data(port);
	int status;
	u8 data = 0;

	status = f81534a_set_register(port, FLOW_CTL_ON, priv->thres_on);
	if (status) {
		dev_err(&port->dev, "%s: set H/W flow control on failed: %d\n",
			__func__, status);
		return;
	}

	status = f81534a_set_register(port, FLOW_CTL_OFF, priv->thres_off);
	if (status) {
		dev_err(&port->dev,
			"%s: set H/W flow control off failed: %d\n", __func__,
			status);
		return;
	}

	//dev_info(&port->dev, "is_hwflow_use_rts: %d\n", priv->is_hwflow_use_rts);

	//en = 0; // disabled by burnintest, rtscts cant used with INTERNAL_LOOPBACK
	if (priv->is_hwflow_use_rts)
		data = en | HW_FLOW_PIN_SEL | HW_FLOW_INVERT; /* RTS */
	else
		data = en | HW_FLOW_INVERT; /* DTR */

#if 1
	// inverted, good
	//dev_info(&port->dev, "Auto H/W flow control inv : %sable\n", en ? "En" : "Dis");

	status = f81534a_set_mask_register(
		port, FLOW_REG,
		HW_FLOW_CONTROL_EN | HW_FLOW_PIN_SEL | HW_FLOW_INVERT, data);
#else
	//dev_info(&port->dev, "Auto H/W flow control: %sable\n", en ? "En" : "Dis");

	status = f81534a_set_mask_register(
		port, FLOW_REG,
		HW_FLOW_CONTROL_EN | HW_FLOW_PIN_SEL | HW_FLOW_INVERT,
		en | HW_FLOW_PIN_SEL);
#endif
	if (status) {
		dev_err(&port->dev, "%s: set H/W flow control failed: %d\n",
			__func__, status);
		return;
	}

	//dev_info(&port->dev, "%s: set flow %d\n", __func__, en);
}

static void __f81534a_set_termios(struct tty_struct *tty,
				  struct usb_serial_port *port,
				  const struct ktermios *old_termios)
{
	struct f81534a_private *priv = usb_get_serial_port_data(port);
	u8 new_lcr = 0;
	int status = 0;
	speed_t baudrate, old_baudrate = 0;

	if (C_BAUD(tty) == B0) {
		f81534a_set_autoflow_mode(port, B0, false);

		f81534a_set_mctrl(port, 0, TIOCM_DTR | TIOCM_RTS);
	} else /* if (old_termios && (old_termios->c_cflag & CBAUD) == B0) */ {
		f81534a_set_autoflow_mode(port, C_BAUD(tty), !!C_CRTSCTS(tty));

		f81534a_set_mctrl(port, TIOCM_DTR | TIOCM_RTS, 0);
	}

	/* Don't change anything if nothing has changed */
	if (old_termios && !tty_termios_hw_change(&tty->termios, old_termios))
		return;

	if (old_termios)
		old_baudrate = old_termios->c_ospeed;
	else
		old_baudrate = F81534A_DEFAULT_BAUD_RATE;

	baudrate = tty_get_baud_rate(tty);
	if (baudrate) {
		if (baudrate > F81534A_MAX_BAUDRATE)
			baudrate = old_baudrate;

		if (old_baudrate != baudrate) {
			status = f81534a_set_baudrate(port, baudrate);
			if (status) {
				baudrate = old_baudrate;
				f81534a_set_baudrate(port, baudrate);
			}

			tty_encode_baud_rate(tty, baudrate, baudrate);

			priv->current_baudrate = baudrate;
		}
	}

	if (priv->force_baudrate) {
		status = f81534a_set_baudrate(port, priv->force_baudrate);
		priv->current_baudrate = priv->force_baudrate;
		dev_info(&port->dev, "%s: force bitrate: %d %d\n", __func__,
			 priv->force_baudrate, status);
	}

	//pr_info("%s: old_baudrate: %d, baudrate: %d\n", __func__, old_baudrate, baudrate);
	if (C_PARENB(tty)) {
		new_lcr |= UART_LCR_PARITY;

		if (!C_PARODD(tty))
			new_lcr |= UART_LCR_EPAR;

		if (C_CMSPAR(tty))
			new_lcr |= UART_LCR_SPAR;
	}

	if (C_CSTOPB(tty))
		new_lcr |= UART_LCR_STOP;

	switch (C_CSIZE(tty)) {
	case CS5:
		new_lcr |= UART_LCR_WLEN5;
		break;
	case CS6:
		new_lcr |= UART_LCR_WLEN6;
		break;
	case CS7:
		new_lcr |= UART_LCR_WLEN7;
		break;
	default:
	case CS8:
		new_lcr |= UART_LCR_WLEN8;
		break;
	}

	mutex_lock(&priv->lock);

	if (priv->is_break)
		new_lcr |= UART_LCR_SBC;

	status = f81534a_set_register(port, LINE_CONTROL_REGISTER, new_lcr);
	if (status) {
		dev_err(&port->dev, "%s failed to set LCR: %d\n", __func__,
			status);
	}

	mutex_unlock(&priv->lock);

	status = f81534a_submit_writer(port, GFP_KERNEL);
	if (status)
		dev_err(&port->dev, "%s: submit failed\n", __func__);
}

static void f81534a_set_termios(struct tty_struct *tty,
				struct usb_serial_port *port,
				struct ktermios *old_termios)
{
	__f81534a_set_termios(tty, port, old_termios);
}

static int f81534a_tiocmget(struct tty_struct *tty)
{
	struct usb_serial_port *port = tty->driver_data;
	struct f81534a_private *priv = usb_get_serial_port_data(port);
	unsigned long flags;
	u8 mcr, msr;
	int r;

	mutex_lock(&priv->lock);
	mcr = priv->modem_control;
	mutex_unlock(&priv->lock);

	spin_lock_irqsave(&port->lock, flags);
	msr = priv->modem_status;
	spin_unlock_irqrestore(&port->lock, flags);

	r = (mcr & UART_MCR_DTR ? TIOCM_DTR : 0) |
	    (mcr & UART_MCR_RTS ? TIOCM_RTS : 0) |
	    (msr & UART_MSR_CTS ? TIOCM_CTS : 0) |
	    (msr & UART_MSR_DCD ? TIOCM_CAR : 0) |
	    (msr & UART_MSR_RI ? TIOCM_RI : 0) |
	    (msr & UART_MSR_DSR ? TIOCM_DSR : 0);

	return r;
}

static int f81534a_tiocmset(struct tty_struct *tty, unsigned int set,
			    unsigned int clear)
{
	struct usb_serial_port *port = tty->driver_data;

	//dev_info(&port->dev, "%s\n", __func__);
	return f81534a_set_mctrl(port, set, clear);
}

static int f81534a_port_init_gpio(struct usb_serial_port *port)
{
	struct f81534a_private *priv = usb_get_serial_port_data(port);
	int status;
	u8 tmp;

	status = f81534a_get_port_register(port, GPIO_REG, 0, 1, &tmp);
	if (status)
		return status;

	status = f81534a_set_port_register(port, GPIO_REG, 0, 1,
					   &priv->gpio_mode);
	if (status)
		return status;

	return 0;
}

static int f81534a_set_rs485(struct usb_serial_port *port)
{
	struct f81534a_private *priv = usb_get_serial_port_data(port);
	struct serial_rs485 rs485;
	int status;
	u8 tmp = 0;

	rs485 = priv->rs485;
	rs485.flags &= SER_RS485_ENABLED | SER_RS485_RTS_ON_SEND;

	if (rs485.flags & SER_RS485_ENABLED) {
		if (rs485.flags & SER_RS485_RTS_ON_SEND) {
			dev_dbg(&port->dev, "%s: uart_mode_rs485_1\n",
				__func__);
			tmp = RS485_EN;
		} else {
			dev_dbg(&port->dev, "%s: uart_mode_rs485\n", __func__);
			tmp = RS485_EN | RS485_INVERT_MODE;
		}
	}

	status = f81534a_set_mask_register(port, MODE_CTRL_REG,
					   RS485_EN | RS485_INVERT_MODE, tmp);
	if (status) {
		mutex_unlock(&priv->lock);
		return status;
	}

	priv->rs485 = rs485;
	return 0;
}

#ifdef USER_MODE
static int f81534a_detect_mode(struct usb_serial_port *port)
{
	struct f81534a_private *priv = usb_get_serial_port_data(port);
	int status;
	u8 tmp[2];
	u8 data;

	status = mutex_lock_killable(&priv->lock);
	if (status)
		goto end;

	/* set port with input mode */
	tmp[0] = GPIO_MODE2_EN | GPIO_MODE1_EN | GPIO_MODE0_EN;

	status = f81534a_set_register(port, GPIO_REG, tmp[0]);
	if (status)
		goto end;

	status = f81534a_get_port_register(port, GPIO_REG, 0, 2, tmp);
	if (status)
		goto end;

	if (tmp[1] == GPIO_ALL_PULL_HIGH) {
		// go back original gpio mode
		dev_dbg(&port->dev, "%s: pin out mode: %x\n", __func__,
			priv->gpio_mode);

		status = f81534a_port_init_gpio(port);
		if (status)
			goto end;
	} else {
		// user mode
		data = tmp[1] & BIT(1);
		data |= (tmp[1] & BIT(2)) >> 2;
		data |= (tmp[1] & BIT(0)) << 2;

		dev_dbg(&port->dev, "%s: user mode: %x %x\n", __func__, tmp[0],
			data);

		memset(&priv->rs485, 0, sizeof(priv->rs485));

		switch (data) {
		case 0:
		case 1:
		case 4:
		case 5:
			dev_dbg(&port->dev, "rs232\n");
			break;

		case 2:
		case 6:
			priv->rs485.flags = SER_RS485_ENABLED;
			dev_dbg(&port->dev, "rs485 inv\n");
			break;

		case 3:
			priv->rs485.flags = SER_RS485_ENABLED |
					    SER_RS485_RTS_ON_SEND;
			dev_dbg(&port->dev, "rs485\n");
			break;

		default:
			status = -EINVAL;
			goto end;
		}

		priv->gpio_mode = tmp[1];
	}

	status = f81534a_set_rs485(port);
	if (status)
		goto end;

end:
	mutex_unlock(&priv->lock);
	return status;
}
#else
static int f81534a_detect_mode(struct usb_serial_port *port)
{
	return 0;
}
#endif

#if MSR_CALLBACK
static void f81534a_msr_callback(struct urb *urb)
{
	struct usb_serial_port *port = urb->context;
	struct f81534a_private *priv = usb_get_serial_port_data(port);
	struct tty_struct *tty;
	unsigned long flags;
	u8 *data = urb->transfer_buffer;
	u8 current_msr;
	int status;

	switch (urb->status) {
	default:
	case 0: /* success */
		break;

	case -ENOENT:
	case -EPIPE:
	case -EPROTO:
	case -ESHUTDOWN:
		return;
	}

	if (urb->actual_length < 1) {
		dev_warn(&port->dev, "%s: length err: %d\n", __func__,
			 urb->actual_length);
		goto resubmit;
	}

	current_msr = data[0];
	//dev_info(&port->dev, "%s: 222 %x\n", __func__, current_msr);

	if (!(current_msr & UART_MSR_ANY_DELTA)) {
		//dev_dbg(&port->dev, "%s: msr not change: %x, old: %x\n", __func__, current_msr, priv->modem_status);
		goto resubmit;
	}

	spin_lock_irqsave(&port->lock, flags);
	priv->modem_status = current_msr;
	spin_unlock_irqrestore(&port->lock, flags);

	dev_dbg(&port->dev, "%s: msr change: from %x to %x\n", __func__,
		priv->modem_status, current_msr);
	dev_dbg(&port->dev, "%s: CTS: %d\n", __func__,
		!!(current_msr & UART_MSR_CTS));

	if (current_msr & UART_MSR_DCTS) {
		port->icount.cts++;

		dev_dbg(&port->dev, "%s: DCTS detect, CTS: %d\n", __func__,
			!!(current_msr & UART_MSR_CTS));

		status = f81534a_submit_writer(port, GFP_KERNEL);
		if (status)
			dev_err(&port->dev, "%s: submit failed\n", __func__);
	}

	if (current_msr & UART_MSR_DDSR)
		port->icount.dsr++;
	if (current_msr & UART_MSR_TERI)
		port->icount.rng++;
	if (current_msr & UART_MSR_DDCD) {
		port->icount.dcd++;
		tty = tty_port_tty_get(&port->port);
		if (tty) {
			usb_serial_handle_dcd_change(
				port, tty, current_msr & UART_MSR_DCD);

			tty_kref_put(tty);
		}
	}

	wake_up_interruptible(&port->port.delta_msr_wait);

resubmit:
	status = usb_submit_urb(priv->msr_urb, GFP_ATOMIC);
	if (status) {
		pr_err("%s: submit write_reg urb failed: %d\n", __func__,
		       status);
	}
}

static int f81534a_msr_async_prepare(struct usb_serial_port *port, gfp_t gfp)
{
	struct f81534a_private *priv = usb_get_serial_port_data(port);
	struct usb_device *dev = port->serial->dev;
	int status = 0;

	priv->dr.bRequestType = F81534A_GET_REGISTER;
	priv->dr.bRequest = F81534A_REGISTER_REQUEST;
	priv->dr.wValue = cpu_to_le16(MODEM_STATUS_REGISTER);
	priv->dr.wIndex = cpu_to_le16(0);
	priv->dr.wLength = cpu_to_le16(1);

	priv->msr_urb = usb_alloc_urb(0, gfp);
	if (!priv->msr_urb)
		return -ENOMEM;

	usb_fill_control_urb(priv->msr_urb, dev, usb_rcvctrlpipe(dev, 0),
			     (void *)&priv->dr, priv->msr_read_buffer, 1,
			     f81534a_msr_callback, port);

	if (priv->is_polling_msr) {
		status = usb_submit_urb(priv->msr_urb, gfp);
		if (status) {
			dev_err(&port->dev,
				"%s: submit write_reg urb failed: %d\n",
				__func__, status);
		}
	}

	return status;
}

static int f81534a_msr_async_clean(struct usb_serial_port *port)
{
	struct f81534a_private *priv = usb_get_serial_port_data(port);

	usb_kill_urb(priv->msr_urb);
	usb_free_urb(priv->msr_urb);

	return 0;
}
#else
static int f81534a_msr_async_prepare(struct usb_serial_port *port, gfp_t gfp)
{
	return 0;
}

static int f81534a_msr_async_clean(struct usb_serial_port *port)
{
	return 0;
}
#endif

static enum hrtimer_restart f81534a_tx_delay_timer_fn(struct hrtimer *t)
{
	struct f81534a_private *priv;
	struct usb_serial_port *port;
	unsigned long flags;
	int status;

	priv = container_of(t, struct f81534a_private, tx_delay_hrt);
	port = priv->port;

	dev_dbg(&port->dev, "%s: 222\n", __func__);

	spin_lock_irqsave(&port->lock, flags);
	set_bit(0, &port->write_urbs_free);
	spin_unlock_irqrestore(&port->lock, flags);

	status = f81534a_submit_writer(port, GFP_ATOMIC);
	if (status) {
		dev_err(&port->dev, "%s: f81534a_submit_writer failed: %d\n",
			__func__, status);
	}

	usb_serial_port_softint(port);

	return HRTIMER_NORESTART;
}

static int f81534a_open(struct tty_struct *tty, struct usb_serial_port *port)
{
	struct f81534a_private *priv = usb_get_serial_port_data(port);
	int result;

	//dev_info(&port->dev, "%s\n", __func__);

#if 1
	result = f81534a_set_mask_register(port, FLOW_REG, UART_RESET,
					   UART_RESET);
	if (result) {
		dev_err(&port->dev, "%s failed to set FLOW_REG: %d\n",
			__func__, result);
		return result;
	}
#endif

	result = f81534a_detect_mode(port);
	if (result)
		return result;

	result = f81534a_get_register(port, MODEM_STATUS_REGISTER,
				      &priv->modem_status);
	if (result)
		return result;

	f81534a_msr_async_prepare(port, GFP_NOIO);

	/* Setup termios */
	if (tty)
		f81534a_set_termios(tty, port, NULL);

	result = f81534a_port_enable(port);
	if (result)
		goto err;

	set_bit(0, &port->write_urbs_free);

	result = usb_serial_generic_open(tty, port);
	if (result)
		goto err;

	return 0;

err:
	f81534a_port_disable(port);

	cancel_work_sync(&priv->interrupt_lsr_work);
	cancel_work_sync(&priv->tx_worker);

	flush_work(&priv->interrupt_lsr_work);
	flush_work(&priv->tx_worker);

	return result;
}

static int f81534a_drain_tx(struct usb_serial_port *port)
{
	int i, r;

	for (i = 0; i < 100; ++i) {
		if (schedule_timeout_interruptible(msecs_to_jiffies(10))) {
			dev_dbg(&port->dev, "%s: killed !!\n", __func__);
			return -EINTR;
		}

		r = f81534a_tx_empty(port);
		if (r)
			return 0;

		dev_dbg(&port->dev, "%s: tx clear pending: %d\n", __func__, i);
	}

	dev_dbg(&port->dev, "%s: tx clear timeout !!\n", __func__);

	return -EINTR;
}

static void f81534a_close(struct usb_serial_port *port)
{
	struct f81534a_private *priv = usb_get_serial_port_data(port);
	f81534a_msr_async_clean(port);

	hrtimer_cancel(&priv->tx_delay_hrt);

	f81534a_drain_tx(port);
	f81534a_port_disable(port);

	usb_serial_generic_close(port);
	//usb_kill_urb(port->interrupt_in_urb);

	cancel_work_sync(&priv->interrupt_lsr_work);
	cancel_work_sync(&priv->tx_worker);

	flush_work(&priv->interrupt_lsr_work);
	flush_work(&priv->tx_worker);

	cancel_delayed_work(&priv->tx_no_active_work);
	flush_delayed_work(&priv->tx_no_active_work);
}

static void f81534a_dtr_rts(struct usb_serial_port *port, int on)
{
	//dev_info(&port->dev, "%s: %d\n", __func__, on);

	if (on) {
		f81534a_set_mctrl(port, TIOCM_DTR | TIOCM_RTS, 0);
	} else {
		f81534a_set_mctrl(port, 0, TIOCM_DTR | TIOCM_RTS);
	}
}

static int f81534a_carrier_raised(struct usb_serial_port *port)
{
	struct f81534a_private *priv = usb_get_serial_port_data(port);
	unsigned long flags;
	u8 msr;

	spin_lock_irqsave(&port->lock, flags);
	msr = priv->modem_status;
	spin_unlock_irqrestore(&port->lock, flags);

	if (msr & UART_MSR_DCD)
		return 1;
	return 0;
}

static bool f81534a_tx_empty(struct usb_serial_port *port)
{
	int status;
	u8 tmp;
	u8 both_empty = UART_LSR_TEMT | UART_LSR_THRE;

	status = f81534a_get_register(port, LINE_STATUS_REGISTER, &tmp);
	if (status) {
		dev_err(&port->dev, "%s: get LSR status < 0\n", __func__);
		return true;
	}

	if ((tmp & both_empty) != both_empty)
		return false;

	return true;
}

static int f81534a_get_serial_info(struct usb_serial_port *port,
				   unsigned long arg)
{
	struct f81534a_private *priv = usb_get_serial_port_data(port);
	struct serial_struct ser;

	memset(&ser, 0, sizeof(ser));

	ser.type = PORT_16550A;
	ser.line = port->minor;
	ser.port = port->port_number;
	ser.baud_base = priv->current_baud_base;

	if (copy_to_user((void __user *)arg, &ser, sizeof(ser)))
		return -EFAULT;

	return 0;
}

static int f81534a_ioctl_set_rs485(struct usb_serial_port *port,
				   struct serial_rs485 __user *arg)
{
	struct f81534a_private *priv = usb_get_serial_port_data(port);
	struct serial_rs485 rs485;
	int status;
	u8 tmp = 0;

	status = copy_from_user(&rs485, (struct serial_rs485 __user *)arg,
				sizeof(rs485));
	if (status)
		return status;

	rs485.flags &= SER_RS485_ENABLED | SER_RS485_RTS_ON_SEND;

	if (mutex_lock_interruptible(&priv->lock))
		return -EINTR;

	if (rs485.flags & SER_RS485_ENABLED) {
		if (rs485.flags & SER_RS485_RTS_ON_SEND) {
			dev_dbg(&port->dev, "%s: uart_mode_rs485_1\n",
				__func__);
			tmp = RS485_EN;
		} else {
			dev_dbg(&port->dev, "%s: uart_mode_rs485\n", __func__);
			tmp = RS485_EN | RS485_INVERT_MODE;
		}
	}

	status = f81534a_set_mask_register(port, MODE_CTRL_REG,
					   RS485_EN | RS485_INVERT_MODE, tmp);
	if (status) {
		mutex_unlock(&priv->lock);
		return status;
	}

	priv->rs485 = rs485;

	mutex_unlock(&priv->lock);
	return 0;
}

static int f81534a_ioctl_get_rs485(struct usb_serial_port *port,
				   struct serial_rs485 __user *arg)
{
	struct f81534a_private *priv = usb_get_serial_port_data(port);
	int status = 0;

	if (mutex_lock_interruptible(&priv->lock))
		return -EINTR;

	if (copy_to_user((struct serial_rs485 *)arg, &priv->rs485,
			 sizeof(priv->rs485)))
		status = -EFAULT;

	mutex_unlock(&priv->lock);
	return status;
}

#ifdef CONFIG_GPIOLIB

static ssize_t pull_high_mode_show(struct device *dev,
				   struct device_attribute *attr, char *buf)
{
	struct usb_interface *intf =
		container_of(dev, struct usb_interface, dev);
	struct usb_device *usb_dev = interface_to_usbdev(intf);
	int status;
	u32 tmp;

	status = f81534a_ctrl_get_register(usb_dev, GPIO_PULL_HIGH_PIN,
					   sizeof(tmp), &tmp);
	if (status)
		return status;

	return sprintf(buf, "%d\n", tmp);
}

static ssize_t pull_high_mode_store(struct device *dev,
				    struct device_attribute *attr,
				    const char *buf, size_t count)
{
	struct usb_interface *intf =
		container_of(dev, struct usb_interface, dev);
	struct usb_device *usb_dev = interface_to_usbdev(intf);
	int status;
	u32 tmp;

	tmp = simple_strtol(buf, NULL, 0);

	status = f81534a_ctrl_set_register(usb_dev, GPIO_PULL_HIGH_PIN,
					   sizeof(tmp), &tmp);
	if (status)
		return status;

	return count;
}

static ssize_t open_drain_mode_show(struct device *dev,
				    struct device_attribute *attr, char *buf)
{
	struct usb_interface *intf =
		container_of(dev, struct usb_interface, dev);
	struct f81534a_ctrl_private *priv = usb_get_intfdata(intf);

	return sprintf(buf, "%lld\n", priv->gpio_opendrain_mode);
}

static ssize_t open_drain_mode_store(struct device *dev,
				     struct device_attribute *attr,
				     const char *buf, size_t count)
{
	struct usb_interface *intf =
		container_of(dev, struct usb_interface, dev);
	struct f81534a_ctrl_private *priv = usb_get_intfdata(intf);

	priv->gpio_opendrain_mode = simple_strtoull(buf, NULL, 0);

	return count;
}

static ssize_t open_drain_dir_show(struct device *dev,
				   struct device_attribute *attr, char *buf)
{
	struct usb_interface *intf =
		container_of(dev, struct usb_interface, dev);
	struct f81534a_ctrl_private *priv = usb_get_intfdata(intf);

	return sprintf(buf, "%lld\n", priv->gpio_opendrain_dir);
}

static ssize_t open_drain_dir_store(struct device *dev,
				    struct device_attribute *attr,
				    const char *buf, size_t count)
{
	struct usb_interface *intf =
		container_of(dev, struct usb_interface, dev);
	struct f81534a_ctrl_private *priv = usb_get_intfdata(intf);

	priv->gpio_opendrain_dir = simple_strtoull(buf, NULL, 0);

	return count;
}

static ssize_t gpio_node_show(struct device *dev,
			      struct device_attribute *attr, char *buf)
{
	struct usb_interface *intf =
		container_of(dev, struct usb_interface, dev);
	struct f81534a_ctrl_private *priv = usb_get_intfdata(intf);

	return sprintf(buf, "%d\n", priv->chip.base);
}

static ssize_t gpio_node_store(struct device *dev,
			       struct device_attribute *attr, const char *buf,
			       size_t count)
{
	return count;
}

static int f81534a_gpio_get(struct gpio_chip *chip, unsigned gpio_num)
{
	struct f81534a_ctrl_private *priv = gpiochip_get_data_fintek(chip);
	struct usb_interface *intf = priv->intf;
	struct usb_device *dev = interface_to_usbdev(intf);
	int status;
	u8 tmp[2];
	int set;
	int idx;

	//dev_info(&intf->dev, "%s\n", __func__);
	set = gpio_num / GPIO_CTRL_MAX_PIN;
	idx = gpio_num % GPIO_CTRL_MAX_PIN;

	status = mutex_lock_killable(&priv->lock);
	if (status) {
		dev_err(&intf->dev, "%s: interrupted!\n", __func__);
		return status;
	}

	status = f81534a_ctrl_get_register(dev, GPIO_CTRL_START + set,
					   sizeof(tmp), tmp);
	if (status) {
		mutex_unlock(&priv->lock);
		return status;
	}

	dev_dbg(&intf->dev, "%s: tmp[1]: %x, current_mode: %d\n", __func__,
		tmp[1], !!(tmp[1] & BIT(idx)));

	mutex_unlock(&priv->lock);

	return !!(tmp[1] & BIT(idx));
}

static int f81534a_gpio_direction_in(struct gpio_chip *chip, unsigned gpio_num)
{
	struct f81534a_ctrl_private *priv = gpiochip_get_data_fintek(chip);
	struct usb_interface *intf = priv->intf;
	struct usb_device *dev = interface_to_usbdev(intf);
	int status;
	int set;
	int idx;
	u8 mask;

	//dev_info(&intf->dev, "%s\n", __func__);
	set = gpio_num / GPIO_CTRL_MAX_PIN;
	idx = gpio_num % GPIO_CTRL_MAX_PIN;
	mask = GPIO_MODE0_EN << idx;

	if (mutex_lock_interruptible(&priv->lock))
		return -EINTR;

	status = f81534a_ctrl_set_mask_register(dev, GPIO_CTRL_START + set,
						mask, mask);
	if (status) {
		mutex_unlock(&priv->lock);
		return status;
	}

	mutex_unlock(&priv->lock);

	return 0;
}

static int f81534a_gpio_direction_out(struct gpio_chip *chip,
				      unsigned gpio_num, int val)
{
	struct f81534a_ctrl_private *priv = gpiochip_get_data_fintek(chip);
	struct usb_interface *intf = priv->intf;
	struct usb_device *dev = interface_to_usbdev(intf);
	int status;
	int set;
	int idx;
	u8 mask;
	u8 data;

	//dev_info(&intf->dev, "%s\n", __func__);
	set = gpio_num / GPIO_CTRL_MAX_PIN;
	idx = gpio_num % GPIO_CTRL_MAX_PIN;
	mask = (GPIO_MODE0_EN << idx) | BIT(idx);
	data = val ? BIT(idx) : 0;

	if (mutex_lock_interruptible(&priv->lock))
		return -EINTR;

	status = f81534a_ctrl_set_mask_register(dev, GPIO_CTRL_START + set,
						mask, data);
	if (status) {
		mutex_unlock(&priv->lock);
		return status;
	}

	mutex_unlock(&priv->lock);

	return 0;
}

static void f81534a_gpio_set(struct gpio_chip *chip, unsigned gpio_num,
			     int val)
{
	//struct f81534a_ctrl_private *priv = gpiochip_get_data_fintek(chip);
	//struct usb_interface *intf = priv->intf;

	//dev_info(&intf->dev, "%s\n", __func__);
	f81534a_gpio_direction_out(chip, gpio_num, val);
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 8, 0)
static int f81534a_gpio_get_direction(struct gpio_chip *chip,
				      unsigned gpio_num)
{
	struct f81534a_ctrl_private *priv = gpiochip_get_data_fintek(chip);
	struct usb_interface *intf = priv->intf;
	struct usb_device *dev = interface_to_usbdev(intf);
	int status;
	u8 tmp[2];
	int set;
	int idx;
	u8 mask;

	set = gpio_num / GPIO_CTRL_MAX_PIN;
	idx = gpio_num % GPIO_CTRL_MAX_PIN;
	mask = GPIO_MODE0_EN << idx;

	//dev_info(&intf->dev, "%s\n", __func__);
	status = mutex_lock_killable(&priv->lock);
	if (status) {
		dev_err(&intf->dev, "%s: interrupted!\n", __func__);
		return status;
	}

	status = f81534a_ctrl_get_register(dev, GPIO_CTRL_START + set,
					   sizeof(tmp), tmp);
	if (status) {
		mutex_unlock(&priv->lock);
		return status;
	}

	//dev_info(&intf->dev, "%s: tmp[1]: %x, current_mode: %d\n", __func__,
	//		tmp[1], current_mode);

	mutex_unlock(&priv->lock);

	if (tmp[0] & mask)
		return GPIOF_DIR_IN;

	return GPIOF_DIR_OUT;
}
#endif

static int f81534a_gpio_request(struct gpio_chip *chip, unsigned offset)
{
	return 0;
}

static void f81534a_gpio_free(struct gpio_chip *chip, unsigned offset)
{
	return;
}

static int f81534a_prepare_gpio(struct usb_interface *intf)
{
	struct f81534a_ctrl_private *priv = usb_get_intfdata(intf);
	struct device_attribute *attr;
	int rc;

	attr = &priv->dev_attr_gpio_pull_mode;
	attr->attr.name = "pull_high_mode";
	attr->attr.mode = VERIFY_OCTAL_PERMISSIONS(0644);
	attr->show = pull_high_mode_show;
	attr->store = pull_high_mode_store;

	rc = device_create_file(&intf->dev, attr);
	if (rc) {
		pr_info("%s: pull_high_mode failed\n", __func__);
		return rc;
	}

	attr = &priv->dev_attr_gpio_operdrain_mode;
	attr->attr.name = "open_drain_mode";
	attr->attr.mode = VERIFY_OCTAL_PERMISSIONS(0644);
	attr->show = open_drain_mode_show;
	attr->store = open_drain_mode_store;

	rc = device_create_file(&intf->dev, attr);
	if (rc) {
		pr_info("%s: open_drain_mode failed\n", __func__);
		device_remove_file(&intf->dev, &priv->dev_attr_gpio_pull_mode);
		return rc;
	}

	attr = &priv->dev_attr_gpio_operdrain_dir;
	attr->attr.name = "open_drain_dir";
	attr->attr.mode = VERIFY_OCTAL_PERMISSIONS(0644);
	attr->show = open_drain_dir_show;
	attr->store = open_drain_dir_store;

	rc = device_create_file(&intf->dev, attr);
	if (rc) {
		pr_info("%s: open_drain_dir failed\n", __func__);
		device_remove_file(&intf->dev,
				   &priv->dev_attr_gpio_operdrain_mode);
		device_remove_file(&intf->dev, &priv->dev_attr_gpio_pull_mode);
		return rc;
	}

	attr = &priv->dev_attr_gpio_node;
	attr->attr.name = "gpiochip_node";
	attr->attr.mode = VERIFY_OCTAL_PERMISSIONS(0644);
	attr->show = gpio_node_show;
	attr->store = gpio_node_store;

	rc = device_create_file(&intf->dev, attr);
	if (rc) {
		pr_info("%s: gpiochip_node failed\n", __func__);
		device_remove_file(&intf->dev,
				   &priv->dev_attr_gpio_operdrain_mode);
		device_remove_file(&intf->dev, &priv->dev_attr_gpio_pull_mode);
		device_remove_file(&intf->dev,
				   &priv->dev_attr_gpio_operdrain_dir);
		return rc;
	}

#if LINUX_VERSION_CODE < KERNEL_VERSION(4, 5, 0)
	priv->chip.dev = &intf->dev;
#else
	priv->chip.parent = &intf->dev;
#endif

	priv->chip.owner = THIS_MODULE;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 8, 0)
	priv->chip.get_direction = f81534a_gpio_get_direction,
#endif
	priv->chip.get = f81534a_gpio_get;
	priv->chip.direction_input = f81534a_gpio_direction_in;
	priv->chip.set = f81534a_gpio_set;
	priv->chip.direction_output = f81534a_gpio_direction_out;
	priv->chip.request = f81534a_gpio_request;
	priv->chip.free = f81534a_gpio_free;
	priv->chip.label = "f81534a";
	priv->chip.ngpio = 3 * 12; /* M0(SD)/M1/M2 */
	priv->chip.base = -1;

	priv->intf = intf;
	mutex_init(&priv->lock);

#if LINUX_VERSION_CODE > KERNEL_VERSION(4, 5, 0)
	rc = devm_gpiochip_add_data(&intf->dev, &priv->chip, priv);
#else
	rc = gpiochip_add(&priv->chip);
#endif
	if (rc) {
		dev_err(&intf->dev, "%s: failed:%d\n", __func__, rc);
		return rc;
	}

	return 0;
}

static int f81534a_release_gpio(struct usb_interface *intf)
{
	struct f81534a_ctrl_private *priv = usb_get_intfdata(intf);

	//devm_gpiochip_remove(&intf->dev, &priv->chip);
#if LINUX_VERSION_CODE > KERNEL_VERSION(4, 5, 0)
	// devm_gpiochip_remove(&intf->dev, &priv->chip);
#else
	gpiochip_remove(&priv->chip);
#endif

	device_remove_file(&intf->dev, &priv->dev_attr_gpio_pull_mode);
	device_remove_file(&intf->dev, &priv->dev_attr_gpio_operdrain_mode);
	device_remove_file(&intf->dev, &priv->dev_attr_gpio_operdrain_dir);
	device_remove_file(&intf->dev, &priv->dev_attr_gpio_node);

	return 0;
}
#else
static int f81534a_prepare_gpio(struct usb_interface *intf)
{
	dev_warn(&intf->dev, "CONFIG_GPIOLIB is not enabled\n");
	dev_warn(&intf->dev, "The GPIOLIB interface will not register\n");

	return 0;
}

static int f81534a_release_gpio(struct usb_interface *intf)
{
	return 0;
}
#endif

#ifdef GPIO_ENABLE
static int f81534a_set_gpio_output_mode(struct usb_serial_port *port,
					int __user *arg)
{
	struct f81534a_private *priv = usb_get_serial_port_data(port);
	int status;
	int value;
	u8 mask = GPIO_MODE2_EN | GPIO_MODE1_EN | GPIO_MODE0_EN;

	if (copy_from_user(&value, (int *)arg, sizeof(value)))
		return -EFAULT;

	if (mutex_lock_interruptible(&priv->lock))
		return -EINTR;

	status = f81534a_set_mask_register(port, GPIO_REG, mask, value & mask);
	if (status) {
		mutex_unlock(&priv->lock);
		return status;
	}

	status = f81534a_get_register(port, GPIO_REG, &priv->gpio_mode);
	if (status) {
		mutex_unlock(&priv->lock);
		return status;
	}

	mutex_unlock(&priv->lock);
	return 0;
}

static int f81534a_set_gpio_output_val(struct usb_serial_port *port,
				       int __user *arg)
{
	struct f81534a_private *priv = usb_get_serial_port_data(port);
	int status;
	int value;
	u8 mask = GPIO_MODE2_OUTPUT | GPIO_MODE1_OUTPUT | GPIO_MODE0_OUTPUT;

	if (copy_from_user(&value, (int *)arg, sizeof(value)))
		return -EFAULT;

	if (mutex_lock_interruptible(&priv->lock))
		return -EINTR;

	status = f81534a_set_mask_register(port, GPIO_REG, mask, value & mask);
	if (status) {
		mutex_unlock(&priv->lock);
		return status;
	}

	status = f81534a_get_register(port, GPIO_REG, &priv->gpio_mode);
	if (status) {
		mutex_unlock(&priv->lock);
		return status;
	}

	mutex_unlock(&priv->lock);

	dev_dbg(&port->dev, "gpio: %x\n", value & mask);
	return 0;
}

static int f81534a_set_gpio_suspend_pull_high(struct usb_serial_port *port,
					      int __user *arg)
{
	struct f81534a_private *priv = usb_get_serial_port_data(port);
	int status;
	int value;
	u8 tmp = 0;

	if (copy_from_user(&value, (int *)arg, sizeof(value)))
		return -EFAULT;

	if (!(value == 0) && !(value == 1))
		return -EINVAL;

	if (mutex_lock_interruptible(&priv->lock))
		return -EINTR;

	if (value)
		tmp = GPIO_SUSPEND_PULL_HIGH;

	status = f81534a_set_mask_register(port, GPIO_REG,
					   GPIO_SUSPEND_PULL_HIGH, tmp);
	if (status) {
		mutex_unlock(&priv->lock);
		return status;
	}

	status = f81534a_get_register(port, GPIO_REG, &priv->gpio_mode);
	if (status) {
		mutex_unlock(&priv->lock);
		return status;
	}

	mutex_unlock(&priv->lock);
	return 0;
}

static int f81534a_get_uart_gpio_mode(struct usb_serial_port *port,
				      int __user *arg)
{
	struct f81534a_private *priv = usb_get_serial_port_data(port);

	if (copy_to_user((int __user *)arg, &priv->gpio_mode,
			 sizeof(priv->gpio_mode)))
		return -EFAULT;

	return 0;
}
#endif

static int f81534a_set_hwflow(struct usb_serial_port *port, int __user *arg)
{
	struct f81534a_private *priv = usb_get_serial_port_data(port);
	struct tty_struct *tty;

	if (copy_from_user(&priv->is_hwflow_use_rts, (int *)arg,
			   sizeof(priv->is_hwflow_use_rts)))
		return -EFAULT;

	tty = tty_port_tty_get(&port->port);
	if (tty) {
		if (C_CRTSCTS(tty))
			f81534a_set_autoflow_mode(port, B0, true);
		else
			f81534a_set_autoflow_mode(port, B0, false);

		tty_kref_put(tty);
	}

	return 0;
}

static int f81534a_get_hwflow(struct usb_serial_port *port, int __user *arg)
{
	struct f81534a_private *priv = usb_get_serial_port_data(port);

	if (copy_to_user((int __user *)arg, &priv->is_hwflow_use_rts,
			 sizeof(priv->is_hwflow_use_rts)))
		return -EFAULT;

	return 0;
}

#ifdef MULTIDROP_ENABLE
static int f81534a_set_9bits(struct usb_serial_port *port, int __user *arg)
{
	struct f81534a_private *priv = usb_get_serial_port_data(port);
	int status;
	u32 mode;
	u8 enable = 0, mask = 0, addr = 0;

	if (copy_from_user(&mode, (int *)arg, sizeof(mode)))
		return -EFAULT;

	if (mode & BIT(17)) {
		enable = 1;
		addr = (mode >> 0) & 0xFF;
		mask = (mode >> 8) & 0xFF;
	}

	if (mutex_lock_interruptible(&priv->lock))
		return -EINTR;

	status = f81534a_set_register(port, SADDR_REG, addr);
	if (status)
		goto end;

	status = f81534a_set_register(port, SADEN_REG, mask);
	if (status)
		goto end;

	status = f81534a_set_mask_register(port, INTERRUPT_ENABLE_REGISTER,
					   BIT(4), BIT(4)); /* clear SM2 */
	if (status)
		goto end;

	status = f81534a_set_mask_register(port, MODE_CONF_REG,
					   AUTO_ADDR_9BITS, AUTO_ADDR_9BITS);
	if (status)
		goto end;

	status = f81534a_set_mask_register(port, MODE_CTRL_REG, ENABLE_9BITS,
					   enable << 4);
	if (status)
		goto end;

	priv->multidrop_flag = mode;
end:
	mutex_unlock(&priv->lock);
	return status;
}

static int f81534a_get_9bits(struct usb_serial_port *port, int __user *arg)
{
	struct f81534a_private *priv = usb_get_serial_port_data(port);
	int status;

	if (mutex_lock_interruptible(&priv->lock))
		return -EINTR;

	if (copy_to_user((int __user *)arg, &priv->multidrop_flag,
			 sizeof(priv->multidrop_flag)))
		status = -EFAULT;
	mutex_unlock(&priv->lock);
	return status;
}

static int f81534a_set_baddr(struct usb_serial_port *port, int __user *arg)
{
	struct f81534a_private *priv = usb_get_serial_port_data(port);
	int status;
	u32 mode;
	u8 enable = 0, addr = 0;

	if (copy_from_user(&mode, (int *)arg, sizeof(mode)))
		return -EFAULT;

	if (mode & BIT(8)) {
		enable = 1;
		addr = (mode >> 0) & 0xFF;
	}

	if (mutex_lock_interruptible(&priv->lock))
		return -EINTR;

	status = f81534a_set_register(port, BADDR_REG, addr);
	if (status)
		goto end;

	status = f81534a_set_mask_register(port, INTERRUPT_ENABLE_REGISTER,
					   BIT(4), BIT(4)); /* clear SM2 */
	if (status)
		goto end;

	status = f81534a_set_mask_register(port, MODE_CONF_REG, BADDR_EN_9BITS,
					   enable << 6);
	if (status)
		goto end;

	priv->baddr_flag = mode;
end:
	mutex_unlock(&priv->lock);
	return status;
}

static int f81534a_get_baddr(struct usb_serial_port *port, int __user *arg)
{
	struct f81534a_private *priv = usb_get_serial_port_data(port);
	int status;

	if (mutex_lock_interruptible(&priv->lock))
		return -EINTR;

	if (copy_to_user((int __user *)arg, &priv->baddr_flag,
			 sizeof(priv->baddr_flag)))
		status = -EFAULT;
	mutex_unlock(&priv->lock);
	return status;
}

#endif

static int f81534a_ioctl(struct tty_struct *tty, unsigned int cmd,
			 unsigned long arg)
{
	struct usb_serial_port *port = tty->driver_data;

	switch (cmd) {
	case TIOCGRS485:
		return f81534a_ioctl_get_rs485(
			port, (struct serial_rs485 __user *)arg);
	case TIOCSRS485:
		return f81534a_ioctl_set_rs485(
			port, (struct serial_rs485 __user *)arg);
#ifdef MULTIDROP_ENABLE
	case FINTEK_SET_MULTI_DROP_MODE:
		return f81534a_set_9bits(port, (int __user *)arg);
	case FINTEK_GET_MULTI_DROP_MODE:
		return f81534a_get_9bits(port, (int __user *)arg);
	case FINTEK_SET_BADDR:
		return f81534a_set_baddr(port, (int __user *)arg);
	case FINTEK_GET_BADDR:
		return f81534a_get_baddr(port, (int __user *)arg);
#endif

#ifdef GPIO_ENABLE
	case FINTEK_SET_UART_OUTPUT_MODE:
		return f81534a_set_gpio_output_mode(port, (int __user *)arg);
	case FINTEK_SET_UART_OUTPUT_VAL:
		return f81534a_set_gpio_output_val(port, (int __user *)arg);
	case FINTEK_SET_UART_SUSPEND_PULL_HIGH:
		return f81534a_set_gpio_suspend_pull_high(port,
							  (int __user *)arg);

	case FINTEK_GET_UART_GPIO_MODE:
		return f81534a_get_uart_gpio_mode(port, (int __user *)arg);

#endif
	case FINTEK_SET_UART_HWFLOW_AS_RTS:
		return f81534a_set_hwflow(port, (int __user *)arg);
	case FINTEK_GET_UART_HWFLOW_AS_RTS:
		return f81534a_get_hwflow(port, (int __user *)arg);

	case TIOCGSERIAL:
		return f81534a_get_serial_info(port, arg);
	default:
		break;
	}
	return -ENOIOCTLCMD;
}

#if MSR_MEASURE_TIME
#define USEC 1000LL
#define MSEC (1000 * USEC)
#define SEC (1000 * MSEC)

static void time_start(struct timespec *start)
{
	getnstimeofday(start);
}

static unsigned long long time_end(struct timespec *start)
{
	long long secs, nsecs;
	struct timespec stop;

	getnstimeofday(&stop);

	secs = stop.tv_sec - start->tv_sec;
	nsecs = stop.tv_nsec - start->tv_nsec;
	if (nsecs < 0) {
		secs--;
		nsecs += SEC;
	}

	return secs * SEC + nsecs;
}
#endif

static void f81534a_interrupt_work(struct work_struct *work)
{
	/*
	struct f81534a_private *priv =
		container_of(work, struct f81534a_private, interrupt_work);
*/
	struct delayed_work *interrupt_work;
	struct f81534a_private *priv;
#if MSR_MEASURE_TIME
	static struct timespec end;
	static unsigned long max_time = 0, time_elapse;
	static int count = 10;
#endif
	interrupt_work = container_of(work, struct delayed_work, work);
	priv = container_of(interrupt_work, struct f81534a_private,
			    interrupt_work);

#if MSR_MEASURE_TIME
	if (count) {
		count--;
	} else {
		time_elapse = time_end(&end);
		if (time_elapse > max_time) {
			max_time = time_elapse;
			dev_info(&priv->port->dev, "%s: time: %lu\n", __func__,
				 max_time);
		}
	}

	time_start(&end);
#endif

	mutex_lock(&priv->lock);
	f81534a_read_msr(priv->port);
	mutex_unlock(&priv->lock);

	if (priv->is_msr_polling /*&& tty_port_initialized(&priv->port->port)*/) {
		dev_dbg(&priv->port->dev, "%s: msr with polling mode\n",
			__func__);
		schedule_delayed_work(&priv->interrupt_work,
				      0 /*usecs_to_jiffies(1)*/);
	}
}

static int f81534a_enable_remote_wakeup(struct usb_serial_port *port)
{
#if 0
	struct usb_device *dev = port->serial->dev;

	return usb_control_msg(dev, usb_sndctrlpipe(dev, 0),
			USB_REQ_SET_FEATURE, USB_RECIP_DEVICE,
			USB_DEVICE_REMOTE_WAKEUP, 0,
			NULL, 0,
			USB_CTRL_SET_TIMEOUT);
#else
	device_set_wakeup_enable(&port->serial->dev->dev, true);

	return 0;
#endif
}

static void f81534a_init_data(struct usb_serial_port *port)
{
	struct f81534a_private *priv = usb_get_serial_port_data(port);

#if 0
	// push-pull high
	priv->gpio_mode = GPIO_MODE2_OUTPUT;	/* reg 0x10e */
#else
	// tri-state with pull-high
	priv->gpio_mode = GPIO_MODE2_EN; /* reg 0x10e */
#endif
	memset(&priv->rs485, 0, sizeof(priv->rs485));
}

static void f81534a_tx_worker(struct work_struct *work)
{
	struct f81534a_private *priv;
	struct usb_serial_port *port;
	struct usb_device *udev;
	struct urb *urb;
	unsigned long flags;
	unsigned char lsr, tx_empty_mask;
	int result, bytes_write;
	int i;

	priv = container_of(work, struct f81534a_private, tx_worker);
	port = priv->port;
	udev = port->serial->dev;
	urb = port->write_urbs[0];

	tx_empty_mask = UART_LSR_TEMT | UART_LSR_THRE;

	//dev_info(&port->dev, "%s: priv->tx_lsr_cnt: %d\n", __func__, priv->tx_lsr_cnt);

	mutex_lock(&priv->tx_lock);

	while (1) {
		for (i = 0; i < priv->tx_lsr_cnt; ++i) {
			result = f81534a_get_register(
				port, LINE_STATUS_REGISTER, &lsr);
			if (result) {
				usb_serial_port_softint(port);
				dev_info(&port->dev, "%s: read lsr failed\n",
					 __func__);
				goto end;
			}

			if ((lsr & tx_empty_mask) != tx_empty_mask) {
				--i;
				continue;
			}
		}

		spin_lock_irqsave(&port->lock, flags);

		if (kfifo_is_empty(&port->write_fifo)) {
			spin_unlock_irqrestore(&port->lock, flags);
			usb_serial_port_softint(port);
			goto end;
		}

#if 0
		urb->transfer_buffer_length = kfifo_out(&port->write_fifo, port->bulk_out_buffers[0],
							priv->tx_size);
#else
		urb->transfer_buffer_length = f81534a_prepare_write_buffer(
			port, port->bulk_out_buffers[0], priv->tx_size);
#endif
		if (!urb->transfer_buffer_length) {
			spin_unlock_irqrestore(&port->lock, flags);
			usb_serial_port_softint(port);
			dev_info(&port->dev,
				 "%s: urb->transfer_buffer_length = 0\n",
				 __func__);
			continue;
		}

		spin_unlock_irqrestore(&port->lock, flags);
retry_tx:
		result = usb_bulk_msg(
			udev,
			usb_sndbulkpipe(udev, port->bulk_out_endpointAddress),
			port->bulk_out_buffers[0], urb->transfer_buffer_length,
			&bytes_write, 10000);
		if (result == ETIMEDOUT) {
			dev_info(&port->dev, "%s: usb_bulk_msg ETIMEDOUT\n",
				 __func__);
			goto retry_tx;
		} else if (result) {
			dev_info(&port->dev, "%s: usb_bulk_msg failed\n",
				 __func__);
			goto end;
		}

		usb_serial_port_softint(port);
	}

end:
	mutex_unlock(&priv->tx_lock);

	cancel_delayed_work(&priv->tx_no_active_work);
	flush_delayed_work(&priv->tx_no_active_work);
	schedule_delayed_work(&priv->tx_no_active_work,
			      msecs_to_jiffies(priv->tx_kfifo_timeout));

	//dev_info(&port->dev, "%s: %d\n", __func__, __LINE__);
}

static int f81534a_port_probe(struct usb_serial_port *port)
{
	struct f81534a_private *priv;
	int status;

	priv = devm_kzalloc(&port->dev, sizeof(*priv), GFP_KERNEL);
	if (!priv)
		return -ENOMEM;

	mutex_init(&priv->lock);
	mutex_init(&priv->tx_lock);
	INIT_DELAYED_WORK(&priv->interrupt_work, f81534a_interrupt_work);
	INIT_DELAYED_WORK(&priv->tx_no_active_work, f81534a_tx_no_active_work);
	INIT_WORK(&priv->interrupt_lsr_work, f81534a_interrupt_lsr_work);
	INIT_WORK(&priv->tx_worker, f81534a_tx_worker);

	priv->port = port;
	priv->is_hwflow_use_rts = true;
	port->port.drain_delay = 256;
	priv->tx_size = 128;
	priv->thres_on = 32;
	priv->thres_off = 96;
	priv->is_polling_msr = true;
	priv->tx_kfifo_timeout = 2000;
	priv->is_random_tx_size = false;

	priv->tx_writer = 1;
	priv->tx_delay = 0; //600;//0;
	priv->tx_delay_type = 1; //0;
	priv->tx_lsr_cnt = 1;

	usb_set_serial_port_data(port, priv);

	hrtimer_init(&priv->tx_delay_hrt, CLOCK_MONOTONIC,
		     HRTIMER_MODE_REL_PINNED);
	priv->tx_delay_hrt.function = f81534a_tx_delay_timer_fn;

	f81534a_init_data(port);

	status = f81534a_port_init_gpio(port);
	if (status)
		return status;

	priv->curr_idx = (port->serial->dev->descriptor.bcdDevice >> 8) & 0xf;
	priv->has_flash =
		!!((port->serial->dev->descriptor.bcdDevice >> 8) & 0x80);

	//dev_info(&port->dev, "%s: curr_idx: %d\n", __func__, priv->curr_idx);

	device_create_file(&port->dev, &dev_attr_f81534a_reg);
	device_create_file(&port->dev, &dev_attr_f81534a_submit);

	device_create_file(&port->dev, &dev_attr_f81534a_tx_size);
	device_create_file(&port->dev, &dev_attr_f81534a_tx_random);
	device_create_file(&port->dev, &dev_attr_f81534a_tx_lsr_cnt);
	device_create_file(&port->dev, &dev_attr_f81534a_tx_writer);
	device_create_file(&port->dev, &dev_attr_f81534a_tx_delay);
	device_create_file(&port->dev, &dev_attr_f81534a_tx_delay_type);
	device_create_file(&port->dev, &dev_attr_f81534a_tx_kfifo);
	device_create_file(&port->dev, &dev_attr_f81534a_tx_kfifo_timeout);

	device_create_file(&port->dev, &dev_attr_f81534a_thres_on);
	device_create_file(&port->dev, &dev_attr_f81534a_thres_off);

	device_create_file(&port->dev, &dev_attr_f81534a_force_baudrate);
	device_create_file(&port->dev, &dev_attr_f81534a_polling_msr);

	if (priv->has_flash)
		dev_dbg(&port->dev, "has flash\n");
	else
		dev_dbg(&port->dev, "not flash present\n");

#if 0
	u8 tmp;
	int i;

	status = f81534a_get_register(port, 0x400, &tmp);
	pr_info("111: %d %x", status, tmp);

	f81534a_set_register(port, 0x400, 0x5a);

	status = f81534a_get_register(port, 0x400, &tmp);
	pr_info("222: %d %x", status, tmp);
#endif
	return f81534a_enable_remote_wakeup(port);
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 12, 0)
static void f81534a_port_remove(struct usb_serial_port *port)
{
	device_remove_file(&port->dev, &dev_attr_f81534a_tx_size);
	device_remove_file(&port->dev, &dev_attr_f81534a_tx_random);
	device_remove_file(&port->dev, &dev_attr_f81534a_tx_lsr_cnt);
	device_remove_file(&port->dev, &dev_attr_f81534a_tx_writer);
	device_remove_file(&port->dev, &dev_attr_f81534a_tx_delay);
	device_remove_file(&port->dev, &dev_attr_f81534a_tx_delay_type);
	device_remove_file(&port->dev, &dev_attr_f81534a_tx_kfifo);
	device_remove_file(&port->dev, &dev_attr_f81534a_tx_kfifo_timeout);

	device_remove_file(&port->dev, &dev_attr_f81534a_thres_on);
	device_remove_file(&port->dev, &dev_attr_f81534a_thres_off);

	device_remove_file(&port->dev, &dev_attr_f81534a_reg);
	device_remove_file(&port->dev, &dev_attr_f81534a_submit);

	device_remove_file(&port->dev, &dev_attr_f81534a_force_baudrate);
	device_remove_file(&port->dev, &dev_attr_f81534a_polling_msr);
}
#else
static int f81534a_port_remove(struct usb_serial_port *port)
{
	device_remove_file(&port->dev, &dev_attr_f81534a_tx_size);
	device_remove_file(&port->dev, &dev_attr_f81534a_tx_random);
	device_remove_file(&port->dev, &dev_attr_f81534a_tx_lsr_cnt);
	device_remove_file(&port->dev, &dev_attr_f81534a_tx_writer);
	device_remove_file(&port->dev, &dev_attr_f81534a_tx_delay);
	device_remove_file(&port->dev, &dev_attr_f81534a_tx_delay_type);
	device_remove_file(&port->dev, &dev_attr_f81534a_tx_kfifo);
	device_remove_file(&port->dev, &dev_attr_f81534a_tx_kfifo_timeout);

	device_remove_file(&port->dev, &dev_attr_f81534a_thres_on);
	device_remove_file(&port->dev, &dev_attr_f81534a_thres_off);

	device_remove_file(&port->dev, &dev_attr_f81534a_reg);
	device_remove_file(&port->dev, &dev_attr_f81534a_submit);

	device_remove_file(&port->dev, &dev_attr_f81534a_force_baudrate);
	device_remove_file(&port->dev, &dev_attr_f81534a_polling_msr);

	return 0;
}
#endif

static int f81534a_suspend(struct usb_serial *serial, pm_message_t message)
{
	struct usb_serial_port *port = serial->port[0];
	struct f81534a_private *priv = usb_get_serial_port_data(port);
	int i;

	hrtimer_cancel(&priv->tx_delay_hrt);

	for (i = 0; i < ARRAY_SIZE(port->read_urbs); ++i)
		usb_kill_urb(port->read_urbs[i]);

	usb_kill_urb(port->interrupt_in_urb);

	if (priv) {
#if MSR_CALLBACK
		usb_kill_urb(priv->msr_urb);
		usb_poison_urb(priv->msr_urb);
#endif

		cancel_work_sync(&priv->interrupt_lsr_work);
		cancel_work_sync(&priv->tx_worker);

		flush_work(&priv->interrupt_lsr_work);
		flush_work(&priv->tx_worker);
	}

	return 0;
}

static int f81534a_resume(struct usb_serial *serial)
{
	struct usb_serial_port *port;
	struct f81534a_private *priv;
	int status;
	u8 tmp;

	port = serial->port[0];
	priv = usb_get_serial_port_data(port);

	status = f81534a_set_register(port, GPIO_REG, priv->gpio_mode);
	if (status)
		return status;

	status = f81534a_get_register(port, GPIO_REG, &tmp);
	if (status)
		return status;

	dev_dbg(&port->dev, "%s: GPIO_REG: %x, gpio_mode: %x\n", __func__, tmp,
		priv->gpio_mode);

	f81534a_enable_remote_wakeup(serial->port[0]);

#if MSR_CALLBACK
	usb_unpoison_urb(priv->msr_urb);
#endif

	if (tty_port_initialized(&port->port)) {
#if MSR_CALLBACK
		status = usb_submit_urb(priv->msr_urb, GFP_NOIO);
		if (status) {
			dev_err(&port->dev,
				"%s: submit write_reg urb failed: %d\n",
				__func__, status);
		}
#endif
		status = usb_serial_generic_submit_read_urbs(port, GFP_NOIO);
		if (status)
			return status;

		status = f81534a_submit_writer(port, GFP_NOIO);
		if (status)
			return status;
	}

	return 0;
}

static int f81534a_write(struct tty_struct *tty, struct usb_serial_port *port,
			 const unsigned char *buf, int count)
{
	//struct f81534a_private *priv = usb_get_serial_port_data(port);
	int result;

	if (!port->bulk_out_size)
		return -ENODEV;

	if (!count)
		return 0;

	count = kfifo_in_locked(&port->write_fifo, buf, count, &port->lock);

	result = f81534a_submit_writer(port, GFP_ATOMIC);
	if (result) {
		dev_err(&port->dev, "%s: f81534a_submit_writer failed: %d\n",
			__func__, result);
		return result;
	}

	return count;
}

static void f81534a_start_hrtimer_us(struct hrtimer *hrt, unsigned long usec)
{
	hrtimer_start(hrt, ns_to_ktime(usec * 1000L), HRTIMER_MODE_REL_PINNED);
}

static void f81534a_write_bulk_callback(struct urb *urb)
{
	struct usb_serial_port *port;
	struct f81534a_private *priv;
	unsigned long tx_delay = 0;

	port = urb->context;
	priv = usb_get_serial_port_data(port);

	if (priv->tx_writer == 2) {
		usb_serial_generic_write_bulk_callback(urb);
		return;
	}

	switch (urb->status) {
	default:
		dev_warn(&port->dev, "%s: URB aborted (%d)\n", __func__,
			 urb->status);
		pr_err("%s: URB aborted: %d\n", __func__, __LINE__);
	case 0: /* success */
		break;

	case -ENOENT:
	case -EPIPE:
	case -EPROTO:
	case -ESHUTDOWN:
		return;
	}

	if (priv->tx_delay_type == 1) {
		tx_delay = priv->tx_delay;

		if (priv->tx_delay == -1) {
			f81534a_tx_delay_timer_fn(&priv->tx_delay_hrt);
			dev_dbg(&port->dev, "%s: %d\n", __func__, __LINE__);
			return;
		}
	} else {
		if (priv->current_baudrate) {
			tx_delay = (10000000UL * urb->transfer_buffer_length +
				    priv->current_baudrate) /
				   priv->current_baudrate;
		}

		tx_delay += priv->tx_delay;
	}

	dev_dbg(&port->dev,
		"%s: 111 tx_delay: %ld, urb->transfer_buffer_length: %d\n",
		__func__, tx_delay, urb->transfer_buffer_length);

	f81534a_start_hrtimer_us(&priv->tx_delay_hrt, tx_delay);
}

static int f81534a_ctrl_generate_ports(struct usb_interface *intf,
				       struct f81534a_device *device)
{
	struct usb_device *dev = interface_to_usbdev(intf);
	uint32_t port_status;
	u8 enable[2];
	u8 tmp;
	u8 mask = GPIO_MODE2_EN | GPIO_MODE1_EN | GPIO_MODE0_EN;
	int status, i;

#if 0
	/* disable pull high */
	enable[0] = 0x00;
	enable[1] = 0x00;
	status = f81534a_ctrl_set_register(dev, GPIO_PULL_HIGH_PIN,
			sizeof(enable), enable);
	if (status) {
		dev_warn(&dev->dev,
				"%s: write GPIO_PULL_HIGH_PIN failed: %d\n",
				__func__, status);
		return status;
	}
#endif
	status = f81534a_ctrl_get_register(dev, F81534A_CMD_ENABLE_PORT,
					   sizeof(enable), enable);

	/* enable ports */
	enable[0] = 0xff;
	enable[1] = 0x8f;

	/* force disable ports */
	if (f81534a_disable_ports) {
		for (i = 0; i < F81534A_MAX_PORT; ++i) {
			if (f81534a_disable_ports & BIT(i)) {
				enable[i / 8] &= ~BIT(i % 8);

				dev_dbg(&dev->dev, "%s: disable port: %d\n",
					__func__, i);
			}
		}

		dev_dbg(&dev->dev, "%s: %xh %xh %xh\n", __func__,
			f81534a_disable_ports, enable[0], enable[1]);
	}

	status = f81534a_ctrl_set_register(dev, F81534A_CMD_ENABLE_PORT,
					   sizeof(enable), enable);
	if (status) {
		dev_warn(&dev->dev,
			 "%s: write F81534A_CMD_ENABLE_PORT failed: %d\n",
			 __func__, status);
		return status;
	}

	/* get port state */
	status = f81534a_ctrl_get_register(dev, F81534A_CMD_PORT_STATUS,
					   sizeof(port_status), &port_status);
	if (status) {
		dev_warn(&dev->dev,
			 "%s: read F81534A_CMD_PORT_STATUS failed: %d\n",
			 __func__, status);
		return status;
	}

#if 0
	pr_info("%x %x %x %x\n", port_status[0], port_status[1], port_status[2], port_status[3]);
	if (port_status[0] != port_status[2])
		return -EAGAIN;

	if ((port_status[1] & 0x0f) != port_status[3])
		return -EAGAIN;
#else
	port_status >>= 16;
	//pr_info("%s: port_status: %x\n", __func__, port_status);
	//pr_info("%s: check_index: %d\n", __func__, device->check_index);
	for (; device->check_index < F81534A_MAX_PORT; ++device->check_index) {
		if (!(port_status & BIT(device->check_index)))
			continue;

		if (f81534a_disable_ports & BIT(device->check_index))
			continue;

		status = f81534a_ctrl_get_register(
			dev, GPIO_CTRL_START + device->check_index,
			sizeof(tmp), &tmp);
		if (status) {
			dev_warn(&dev->dev,
				 "%s: read GPIO_CTRL_START failed: %d\n",
				 __func__, status);
			return status;
		}

		//pr_info("%s: tmp: %x, mask: %x\n", __func__, tmp, mask);
		if ((tmp & mask) == mask)
			break;
	}

	if (device->check_index < F81534A_MAX_PORT)
		return -EAGAIN;
#endif

	return 0;
}

static void f81534a_ctrl_dwork_func(struct work_struct *work)
{
	struct f81534a_device *device;
	int status;

	//pr_info("%s\n", __func__);

	mutex_lock(&device_mutex);
	list_for_each_entry(device, &device_list, list) {
		if (device->check_index >= F81534A_MAX_PORT)
			continue;

		//pr_info("device: %x\n", device);
		if (device->check_retry >= F81534A_CTRL_RETRY)
			continue;

		++device->check_retry;

		status = f81534a_ctrl_generate_ports(device->intf, device);
		if (!status) {
			/* make this device generated */
			device->check_index = F81534A_MAX_PORT;

			pr_info("%s: generated complete\n", __func__);
		} else if (status == -EAGAIN) {
			pr_info("%s: delayed generating: %d\n", __func__,
				device->check_retry);

			schedule_delayed_work(
				&f81534a_ctrl_dwork,
				msecs_to_jiffies(F81534A_CTRL_TIMER));
			break;
		} else {
			/* skip this device to generate */
			device->check_index = F81534A_MAX_PORT;

			pr_err("%s: error: %d, do next device gen\n", __func__,
			       status);
		}
	}

	//pr_info("device_list: %x", &device_list);
	mutex_unlock(&device_mutex);
}

static int f81534a_ctrl_probe(struct usb_interface *intf,
			      const struct usb_device_id *id)
{
	struct usb_device *dev = interface_to_usbdev(intf);
	struct f81534a_ctrl_private *priv;
	struct f81534a_device *device;
	int status;
	//u8 buf[2] = {0};

	priv = devm_kzalloc(&intf->dev, sizeof(*priv), GFP_KERNEL);
	if (!priv)
		return -ENOMEM;

	device = devm_kzalloc(&intf->dev, sizeof(*device), GFP_KERNEL);
	if (!device)
		return -ENOMEM;

	mutex_init(&priv->lock);
	usb_set_intfdata(intf, priv);

	status = f81534a_prepare_gpio(intf);
	if (status)
		return status;

	INIT_LIST_HEAD(&device->list);
	device->intf = intf;

	mutex_lock(&device_mutex);
	list_add_tail(&device->list, &device_list);
	mutex_unlock(&device_mutex);

	dev = usb_get_dev(dev);

#if 0
	pr_info("--- fintek debug, cut from here ---\n");
	status = f81534a_ctrl_get_register_raw(dev, 0x1600, 1, &buf[0]);
	pr_info("read 1600h 1byte: %x, %d\n", buf[0], status);

	status = f81534a_ctrl_set_register_raw(dev, 0x1600, 1, &buf[0]);
	pr_info("write 1600h 1byte: %x, %d\n", buf[0], status);

	status = f81534a_ctrl_get_register_raw(dev, F81534A_CMD_PORT_STATUS, 2, buf);
	pr_info("read 117h 2byte: %x %x, %d\n", buf[0], buf[1], status);

	status = f81534a_ctrl_get_register_raw(dev, F81534A_CMD_ENABLE_PORT, 2, buf);
	pr_info("read 116h 2byte: %x %x, %d\n", buf[0], buf[1], status);

	buf[0] = 0xff;
	buf[1] = 0x8f;
	status = f81534a_ctrl_set_register_raw(dev, F81534A_CMD_ENABLE_PORT, 2, buf);
	pr_info("write 116h 2byte: %x %x, %d\n", buf[0], buf[1], status);

	pr_info("--- fintek debug, cut end ---\n");
#else
	schedule_delayed_work(&f81534a_ctrl_dwork,
			      msecs_to_jiffies(F81534A_CTRL_TIMER));
#endif
	device_create_file(&intf->dev, &dev_attr_f81534a_id);

	return 0;
}

static void f81534a_ctrl_disconnect(struct usb_interface *intf)
{
	struct f81534a_ctrl_private *priv;
	struct f81534a_device *device;
	struct usb_device *dev;
	u8 enable[2];

	mutex_lock(&device_mutex);

	list_for_each_entry(device, &device_list, list) {
		if (device->intf == intf) {
			list_del(&device->list);

			priv = usb_get_intfdata(intf);
			dev = interface_to_usbdev(intf);

			mutex_lock(&priv->lock);
			f81534a_release_gpio(intf);
			mutex_unlock(&priv->lock);

			usb_put_dev(dev);

			device_remove_file(&intf->dev, &dev_attr_f81534a_id);
			break;
		}
	}

	mutex_unlock(&device_mutex);

	/* disable ports */
	enable[0] = 0x00;
	enable[1] = 0x80;

	f81534a_ctrl_set_register(dev, F81534A_CMD_ENABLE_PORT, sizeof(enable),
				  enable);
}

#if 0
static int f81534a_ctrl_probe_old(struct usb_interface *intf,
				const struct usb_device_id *id)
{
	struct usb_device *dev = interface_to_usbdev(intf);
	struct device *device;
	struct f81534a_ctrl_private *priv;
	u8 enable[4], port_status[4];
	u8 path[128];
	int status;
	int retry = 40;
	int first_idx = -1;


	pr_info("gpio device\n");

	status = snprintf(path, sizeof(path) - 1, "%d-%s", dev->bus->busnum,
				dev->devpath);
	if (status >= sizeof(path) - 1)
		return -ENOMEM;

	pr_info("path: %s\n", path);

	dev = usb_get_dev(dev);
	mutex_lock(&device_mutex);

	status = f81534a_ctrl_check_stable(intf);
	if (status)
		goto out;

	priv = devm_kzalloc(&intf->dev, sizeof(*priv), GFP_KERNEL);
	if (!priv) {
		status = -ENOMEM;
		goto out;
	}

	first_idx = ffz(device_map);
	device = device_create(f81534_ctrl_class, NULL,
				MKDEV(driver_major, first_idx), NULL,
				F81534A_CTRL_DEVICE, path);
	if (PTR_ERR_OR_ZERO(device)) {
		status = PTR_ERR_OR_ZERO(device);
		goto out;
	}

	set_bit(first_idx, &device_map);

#if GPIO_DEFAULT_RS232
	status = f81534a_init_gpio(intf);
	if (status)
		goto out;
#endif

	enable[0] = 0xff;
	enable[1] = 0x8f;

	status = f81534a_ctrl_set_register(dev, F81534A_CMD_ENABLE_PORT,
			sizeof(enable), enable);
	if (status) {
		dev_warn(&dev->dev,
				"%s: write F81534A_CMD_ENABLE_PORT 1 failed: %d\n",
				__func__, status);
		goto out;
	}
#if 1
	retry = 40;
	while (retry--) {
		msleep(500);

		status = f81534a_ctrl_get_register(dev,
				F81534A_CMD_PORT_STATUS, sizeof(port_status),
				port_status);
		if (status) {
			dev_warn(&dev->dev,
					"%s: read F81534A_CMD_PORT_STATUS failed: %d\n",
					__func__, status);
			goto out;
		}	

		pr_info("%x %x %x %x\n", port_status[0], port_status[1], port_status[2], port_status[3]);
		if (port_status[0] != port_status[2])
			continue;

		if ((port_status[1] & 0x0f) != port_status[3])
			continue;	

		break;
	}
#endif

	priv->device_idx = first_idx;
	usb_set_intfdata(intf, priv);

	status = f81534a_prepare_gpio(intf);
out:
	mutex_unlock(&device_mutex);
	pr_info("%s: check finish\n", __func__);

	if (status) {
		if (first_idx != -1) {
			device_destroy(f81534_ctrl_class, MKDEV(driver_major, first_idx));
			clear_bit(first_idx, &device_map);
		}
		usb_put_dev(dev);

		pr_info("%s: status: %d\n", __func__, status);
		return status;
	}

	return 0;
}

static void f81534a_ctrl_disconnect_old(struct usb_interface *intf)
{
	struct usb_device *dev = interface_to_usbdev(intf);
	struct f81534a_ctrl_private *priv = usb_get_intfdata(intf);

	f81534a_release_gpio(intf);

	device_destroy(f81534_ctrl_class, MKDEV(driver_major, priv->device_idx));
	clear_bit(priv->device_idx, &device_map);

	usb_put_dev(dev);
}
#endif

static int f81534a_ctrl_suspend(struct usb_interface *intf,
				pm_message_t message)
{
	struct f81534a_device *device;

	flush_delayed_work(&f81534a_ctrl_dwork);

	mutex_lock(&device_mutex);

	list_for_each_entry(device, &device_list, list) {
		device->check_index = 0;
		device->check_retry = 0;
	}

	mutex_unlock(&device_mutex);

	return 0;
}

static int f81534a_ctrl_resume(struct usb_interface *intf)
{
	schedule_delayed_work(&f81534a_ctrl_dwork,
			      msecs_to_jiffies(F81534A_CTRL_TIMER));

	return 0;
}

static struct usb_serial_driver f81534a_device = {
	.driver = {
		.owner =	THIS_MODULE,
		.name =		"f81534a_uart"DRIVER_NAME_APPEND,
	},
	.id_table =		f81534a_uart_id_table,
	.num_ports =		1,
#if 0
	.bulk_in_size =		256,
	.bulk_out_size =	256, /* 8 */
#endif
	//.bulk_out_size =	64, /* 8 */
	.write =		f81534a_write,
	.write_bulk_callback =	f81534a_write_bulk_callback,
	.prepare_write_buffer = f81534a_prepare_write_buffer_locked, //f81534a_prepare_write_buffer,
	.open =			f81534a_open,
	.close =		f81534a_close,
	.dtr_rts =		f81534a_dtr_rts,
	.carrier_raised =	f81534a_carrier_raised,
	.ioctl =		f81534a_ioctl,
	.break_ctl =		f81534a_break_ctl,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 1, 0)
	.set_termios =		__f81534a_set_termios,
#else
	.set_termios =		f81534a_set_termios,
#endif
	.tx_empty =		f81534a_tx_empty,
	.tiocmget =		f81534a_tiocmget,
	.tiocmset =		f81534a_tiocmset,
	.tiocmiwait =		usb_serial_generic_tiocmiwait,
	.process_read_urb =	f81534a_process_read_urb,
	.read_int_callback =	f81534a_read_int_callback,
	.port_probe =		f81534a_port_probe,
	.port_remove =		f81534a_port_remove,
	.suspend =		f81534a_suspend,
	.resume =		f81534a_resume,
};

static struct usb_serial_driver *const f81534a_serial_drivers[] = {
	&f81534a_device,
	&f81232_device,
	NULL,
};

static struct usb_driver f81534a_ctrl_driver = {
	.name = "f81534a_ctrl" DRIVER_NAME_APPEND,
	.probe = f81534a_ctrl_probe,
	.disconnect = f81534a_ctrl_disconnect,
	.suspend = f81534a_ctrl_suspend,
	.resume = f81534a_ctrl_resume,
	.id_table = f81534a_ctrl_id_table
};

static int __init f81534a_init(void)
{
	int status;

	pr_info("Fintek F81232/F81532A/534A/535/536 Driver version: %s\n",
		VERSION);

	INIT_DELAYED_WORK(&f81534a_ctrl_dwork, f81534a_ctrl_dwork_func);

	status = usb_register_driver(&f81534a_ctrl_driver, THIS_MODULE,
				     KBUILD_MODNAME);
	if (status) {
		pr_info("f81534a_ctrl_driver register failed: %d\n", status);
		return status;
	}

	status = usb_serial_register_drivers(
		f81534a_serial_drivers, KBUILD_MODNAME, all_uart_id_table);
	if (status) {
		pr_info("f81534a_serial_drivers register failed: %d\n",
			status);
		usb_deregister(&f81534a_ctrl_driver);
		return status;
	}

	return 0;
}

static void __exit f81534a_exit(void)
{
	flush_delayed_work(&f81534a_ctrl_dwork);

	usb_serial_deregister_drivers(f81534a_serial_drivers);
	usb_deregister(&f81534a_ctrl_driver);
}

module_init(f81534a_init);
module_exit(f81534a_exit);

MODULE_DESCRIPTION("Fintek F81232/F81534A USB to serial adaptor driver");
MODULE_AUTHOR("Greg Kroah-Hartman <gregkh@linuxfoundation.org>");
MODULE_AUTHOR("Peter Hong <peter_hong@fintek.com.tw>");
MODULE_LICENSE("GPL v2");
